import contextlib
import torch

def set_default_option(defopt, opt=None):
    if opt is None:
        opt = {}
    defopt.update(opt)
    return defopt

def extract_differentiable_tensors(model):
    res = list(model.parameters())

    # add differentiable and non-parameters attribute in the model
    for varname in model.__dict__:
        var = model.__dict__[varname]
        if not isinstance(var, torch.Tensor): continue
        if var.requires_grad:
            res.append(var)
    return res

@contextlib.contextmanager
def dummy_context_manager():
    yield None

class TensorNonTensorSeparator(object):
    """
    Class that provides function to separate/combine tensors and nontensors
    parameters.
    """
    def __init__(self, params, varonly=True):
        """
        Params is a list of tensor or non-tensor to be splitted into
        tensor/non-tensor
        """
        self.tensor_idxs = []
        self.tensor_params = []
        self.nontensor_idxs = []
        self.nontensor_params = []
        self.nparams = len(params)
        for (i,p) in enumerate(params):
            if isinstance(p, torch.Tensor) and ((varonly and p.requires_grad) or (not varonly)):
                self.tensor_idxs.append(i)
                self.tensor_params.append(p)
            else:
                self.nontensor_idxs.append(i)
                self.nontensor_params.append(p)
        self.alltensors = len(self.tensor_idxs) == self.nparams

    def get_tensor_params(self):
        return self.tensor_params

    def get_nontensor_params(self):
        return self.nontensor_params

    def get_tensor_idxs(self):
        return self.tensor_idxs

    def get_nontensor_idxs(self):
        return self.nontensor_idxs

    def ntensors(self):
        return len(self.tensor_idxs)

    def nnontensors(self):
        return len(self.nontensor_idxs)

    def reconstruct_params(self, tensor_params, nontensor_params=None):
        if nontensor_params is None:
            nontensor_params = self.nontensor_params
        if len(tensor_params) + len(nontensor_params) != self.nparams:
            raise ValueError("The total length of tensor and nontensor params "\
                "do not match with the expected length: %d instead of %d" % \
                (len(tensor_params)+len(nontensor_params), self.nparams))
        if self.alltensors:
            return tensor_params

        params = [None for _ in range(self.nparams)]
        for nidx,p in zip(self.nontensor_idxs, nontensor_params):
            params[nidx] = p
        for idx,p in zip(self.tensor_idxs, tensor_params):
            params[idx] = p
        return params

class TensorPacker(object):
    def __init__(self, tensors):
        self.idx_shapes = []
        istart = 0
        for i,p in enumerate(tensors):
            ifinish = istart + torch.numel(p)
            self.idx_shapes.append((istart, ifinish, p.shape))
            istart = ifinish

    def flatten(self, y_list):
        return torch.cat([y.reshape(-1) for y in y_list], dim=-1)

    def pack(self, y):
        yshapem1 = y.shape[:-1]
        return tuple([y[...,istart:ifinish].reshape((*yshapem1, *shape)) for (istart,ifinish,shape) in self.idx_shapes])



import sys
import inspect
import warnings
from abc import abstractmethod
from contextlib import contextmanager
import copy
import traceback as tb
import torch
from typing import Sequence, Union, Mapping, Sequence, Dict, List
from xitorch._utils.exceptions import GetSetParamsError
from xitorch._utils.attr import get_attr, set_attr, del_attr

__all__ = ["EditableModule"]

torch_float_type = [torch.float32, torch.float, torch.float64, torch.float16]

class EditableModule(object):
    """
    ``EditableModule`` is a base class to enable classes that it inherits be
    converted to pure functions for higher order derivatives purpose.
    """
    def getparams(self, methodname:str) -> Sequence[torch.Tensor]:
        # Returns a list of tensor parameters used in the object's operations

        paramnames = self.getparamnames(methodname)
        return [get_attr(self, name) for name in paramnames]

    def setparams(self, methodname:str, *params) -> int:
        # Set the input parameters to the object's parameters to make a copy of
        # the operations.
        # *params is an excessive list of the parameters to be set and the
        # method will return the number of parameters it sets.
        paramnames = self.getparamnames(methodname)
        for name,val in zip(paramnames, params):
            set_attr(self, name, val)
        return len(params)

    @abstractmethod
    def getparamnames(self, methodname:str, prefix:str="") -> List[str]:
        """
        This method should list tensor names that affect the output of the
        method with name indicated in ``methodname``.
        If the ``methodname`` is not on the list in this function, it should
        raise ``KeyError``.

        Arguments
        ---------
        methodname: str
            The name of the method of the class.
        prefix: str
            The prefix to be appended in front of the parameters name.
            This usually contains the dots.

        Returns
        -------
        Sequence of string
            Sequence of name of parameters affecting the output of the method.

        Raises
        ------
        KeyError
            If the list in this function does not contain ``methodname``.

        Example
        -------
        .. testsetup::

            import torch
            import xitorch

        .. doctest::

            >>> class A(xitorch.EditableModule):
            ...     def __init__(self, a):
            ...         self.b = a*a
            ...
            ...     def mult(self, x):
            ...         return self.b * x
            ...
            ...     def getparamnames(self, methodname, prefix=""):
            ...         if methodname == "mult":
            ...             return [prefix+"b"]
            ...         else:
            ...             raise KeyError()
        """
        pass

    def getuniqueparams(self, methodname:str) -> List[torch.Tensor]:
        allparams = self.getparams(methodname)
        idxs = self._get_unique_params_idxs(methodname, allparams)
        return [allparams[i] for i in idxs]

    def setuniqueparams(self, methodname:str, *uniqueparams) -> int:
        nparams = self._number_of_params[methodname]
        allparams = [None for _ in range(nparams)]
        maps = self._unique_params_maps[methodname]

        for j in range(len(uniqueparams)):
            jmap = maps[j]
            p = uniqueparams[j]
            for i in jmap:
                allparams[i] = p

        return self.setparams(methodname, *allparams)

    def _get_unique_params_idxs(self, methodname:str,
            allparams:Union[Sequence[torch.Tensor],None]=None) -> Sequence[int]:

        if not hasattr(self, "_unique_params_idxs"):
            self._unique_params_idxs = {} # type: Dict[str,Sequence[int]]
            self._unique_params_maps = {}
            self._number_of_params = {}

        if methodname in self._unique_params_idxs:
            return self._unique_params_idxs[methodname]
        if allparams is None:
            allparams = self.getparams(methodname)

        # get the unique ids
        ids = [] # type: List[int]
        idxs = []
        idx_map = [] # type: List[List[int]]
        for i in range(len(allparams)):
            param = allparams[i]
            id_param = id(param)

            # search the id if it has been added to the list
            try:
                jfound = ids.index(id_param)
                idx_map[jfound].append(i)
                continue
            except ValueError:
                pass

            ids.append(id_param)
            idxs.append(i)
            idx_map.append([i])

        self._number_of_params[methodname] = len(allparams)
        self._unique_params_idxs[methodname] = idxs
        self._unique_params_maps[methodname] = idx_map
        return idxs

    @contextmanager
    def useuniqueparams(self, methodname:str, *params):
        try:
            _orig_params_ = self.getuniqueparams(methodname)
            identical = _all_equal(_orig_params_, params)
            if not identical:
                self.setuniqueparams(methodname, *params)
            yield self
        except Exception as exc:
            raise exc
        finally:
            if not identical:
                self.setuniqueparams(methodname, *_orig_params_)

    ############# debugging #############
    def assertparams(self, method, *args, **kwargs):
        """
        Perform a rigorous check on the implemented ``getparamnames``
        in the class for a given method and its arguments as well as keyword
        arguments.
        It raises warnings if there are missing or excess parameters in the
        ``getparamnames`` implementation.

        Arguments
        ---------
        method: callable method
            The method of this class to be tested
        *args:
            Arguments of the method
        **kwargs:
            Keyword arguments of the method

        Example
        -------
        .. testsetup:: assertparams

            import torch
            import xitorch
            import sys
            sys.stderr = sys.stdout

        .. doctest:: assertparams

            >>> class AClass(xitorch.EditableModule):
            ...     def __init__(self, a):
            ...         self.a = a
            ...         self.b = a*a
            ...
            ...     def mult(self, x):
            ...         return self.b * x
            ...
            ...     def getparamnames(self, methodname, prefix=""):
            ...         if methodname == "mult":
            ...             return [prefix+"a"]  # intentionally wrong
            ...         else:
            ...             raise KeyError()
            >>> a = torch.tensor(2.0).requires_grad_()
            >>> x = torch.tensor(0.4).requires_grad_()
            >>> A = AClass(a)
            >>> A.assertparams(A.mult, x) # doctest:+ELLIPSIS
            <doctest assertparams[4]>:1: UserWarning: getparams for AClass.mult does not include: b
              A.assertparams(A.mult, x) # doctest:+ELLIPSIS
            <doctest assertparams[4]>:1: UserWarning: getparams for AClass.mult has excess parameters: a
              A.assertparams(A.mult, x) # doctest:+ELLIPSIS
            "mult" method check done
        """
        # check the method input
        if not inspect.ismethod(method):
            raise TypeError("The input method must be a method")
        methodself = method.__self__
        if methodself is not self:
            raise RuntimeError("The method does not belong to the same instance")

        methodname = method.__name__

        self.__assert_method_preserve(method, *args, **kwargs) # assert if the method preserve the float tensors of the object
        self.__assert_match_getsetparams(methodname) # check if getparams and setparams matched the tensors
        self.__assert_get_correct_params(method, *args, **kwargs) # check if getparams returns the correct tensors
        print('"%s" method check done' % methodname)

    def __assert_method_preserve(self, method, *args, **kwargs):
        # this method assert if method does not change the float tensor parameters
        # of the object (i.e. it preserves the state of the object)

        all_params0, names0 = _get_tensors(self)
        all_params0 = [p.clone() for p in all_params0]
        method(*args, **kwargs)
        all_params1, names1 = _get_tensors(self)

        # now assert if all_params0 == all_params1
        clsname = method.__self__.__class__.__name__
        methodname = method.__name__
        msg = "The method %s.%s does not preserve the object's float tensors" % (clsname, methodname)
        if len(all_params0) != len(all_params1):
            raise GetSetParamsError(msg)

        for p0,p1 in zip(all_params0, all_params1):
            if p0.shape != p1.shape:
                raise GetSetParamsError(msg)
            if not torch.allclose(p0,p1):
                raise GetSetParamsError(msg)

    def __assert_match_getsetparams(self, methodname):
        # this function assert if get & set params functions correspond to the
        # same parameters in the same order

        # count the number of parameters in getparams and setparams
        params0 = self.getparams(methodname)
        len_setparams0 = self.setparams(methodname, *params0)
        if len_setparams0 != len(params0):
            raise GetSetParamsError("The number of parameters returned by getparams and set by setparams do not match \n"\
                "(getparams: %d, setparams: %d)" % (len(params0), len_setparams0))

        # check if the params are assigned correctly in the correct order
        params1 = self.getparams(methodname)
        names1 = self.getparamnames(methodname)
        for i,p0,p1 in zip(range(len(params0)), params0, params1):
            if id(p0) != id(p1):
                msg = "The parameter %s in getparams and setparams does not match" % names1[i]
                raise GetSetParamsError(msg)

    def __assert_get_correct_params(self, method, *args, **kwargs):
        # this function perform checks if the getparams on the method returns
        # the correct tensors

        methodname = method.__name__
        clsname = method.__self__.__class__.__name__

        # get all tensor parameters in the object
        all_params, all_names = _get_tensors(self)
        def _get_tensor_name(param):
            for i in range(len(all_params)):
                if id(all_params[i]) == id(param):
                    return all_names[i]
            return None

        # get the parameter tensors used in the operation and the tensors specified by the developer
        oper_names, oper_params = self.__list_operating_params(method, *args, **kwargs)
        user_names = self.getparamnames(method.__name__)
        user_params = [get_attr(self, name) for name in user_names]
        user_params_id = [id(p) for p in user_params]
        oper_params_id = [id(p) for p in oper_params]
        user_params_id_set = set(user_params_id)
        oper_params_id_set = set(oper_params_id)


        # check if the userparams contains non-tensor
        for i in range(len(user_params)):
            param = user_params[i]
            if (not isinstance(param, torch.Tensor)) or (isinstance(param, torch.Tensor) and param.dtype not in torch_float_type):
                msg = "Parameter %s is a non-floating point tensor" % user_names[i]
                raise GetSetParamsError(msg)

        # check if there are missing parameters (present in operating params, but not in the user params)
        missing_names = []
        for i in range(len(oper_names)):
            if oper_params_id[i] not in user_params_id_set:
            # if oper_names[i] not in user_names:
                missing_names.append(oper_names[i])
        # if there are missing parameters, give a warning (because the program
        # can still run correctly, e.g. missing parameters are parameters that
        # are never set to require grad)
        if len(missing_names) > 0:
            msg = "getparams for %s.%s does not include: %s" % (clsname, methodname, ", ".join(missing_names))
            warnings.warn(msg, stacklevel=3)

        # check if there are excessive parameters (present in the user params, but not in the operating params)
        excess_names = []
        for i in range(len(user_names)):
            if user_params_id[i] not in oper_params_id_set:
            # if user_names[i] not in oper_names:
                excess_names.append(user_names[i])
        # if there are excess parameters, give warnings
        if len(excess_names) > 0:
            msg = "getparams for %s.%s has excess parameters: %s" % \
                (clsname, methodname, ", ".join(excess_names))
            warnings.warn(msg, stacklevel=3)

    def __list_operating_params(self, method, *args, **kwargs):
        # Sequence the tensors used in executing the method by calling the method
        # and see which parameters are connected in the backward graph

        # get all the tensors recursively
        all_tensors, all_names = _get_tensors(self)

        # copy the tensors and require them to be differentiable
        copy_tensors0 = [tensor.clone().detach().requires_grad_() for tensor in all_tensors]
        copy_tensors = copy.copy(copy_tensors0)
        _set_tensors(self, copy_tensors)

        # run the method and see which one has the gradients
        output = method(*args, **kwargs).sum()
        grad_tensors = torch.autograd.grad(output, copy_tensors0, retain_graph=True, allow_unused=True)

        # return the original tensor
        all_tensors_copy = copy.copy(all_tensors)
        _set_tensors(self, all_tensors_copy)

        names = []
        params = []
        for i, grad in enumerate(grad_tensors):
            if grad is None:
                continue
            names.append(all_names[i])
            params.append(all_tensors[i])

        return names, params

def getmethodparams(method):
    if not inspect.ismethod(method):
        return []
    obj = method.__self__
    methodname = method.__name__
    if not isinstance(obj, EditableModule):
        return []
    return obj.getparams(methodname)

def setmethodparams(method, *params):
    if not inspect.ismethod(method):
        return
    obj = method.__self__
    methodname = method.__name__
    if not isinstance(obj, EditableModule):
        return 0
    return obj.setparams(methodname, *params)

def _all_equal(params1, params2):
    for p1, p2 in zip(params1, params2):
        if id(p1) != id(p2):
            return False
    return True

############################ traversing functions ############################
def _traverse_obj(obj, prefix, action, crit, max_depth=20, exception_ids=None):
    """
    Traverse an object to get/set variables that are accessible through the object.
    """
    if exception_ids is None:
        # None is set as default arg to avoid expanding list for multiple
        # invokes of _get_tensors without exception_ids argument
        exception_ids = set()

    if hasattr(obj, "__dict__"):
        generator = obj.__dict__.items()
        name_format = "{prefix}{key}"
        objdict = obj.__dict__
    elif hasattr(obj, "__iter__"):
        generator = obj.items() if isinstance(obj, dict) else enumerate(obj)
        name_format = "{prefix}[{key}]"
        objdict = obj
    else:
        raise RuntimeError("The object must be iterable or keyable")

    for key,elmt in generator:
        name = name_format.format(prefix=prefix, key=key)
        if crit(elmt):
            action(elmt, name, objdict, key)
            continue

        hasdict = hasattr(elmt, "__dict__")
        hasiter = hasattr(elmt, "__iter__")
        if hasdict or hasiter:
            # add exception to avoid infinite loop if there is a mutual dependant on objects
            if id(elmt) in exception_ids:
                continue
            else:
                exception_ids.add(id(elmt))

            if max_depth > 0:
                _traverse_obj(elmt,
                    action = action,
                    crit = crit,
                    prefix = name+"." if hasdict else name,
                    max_depth = max_depth - 1,
                    exception_ids=exception_ids)
            else:
                raise RecursionError("Maximum number of recursion reached")

def _get_tensors(obj, prefix="", max_depth=20):
    """
    Collect all tensors in an object recursively and return the tensors as well
    as their "names" (names meaning the address, e.g. "self.a[0].elmt").

    Arguments
    ---------
    * obj: an instance
        The object user wants to traverse down
    * prefix: str
        Prefix of the name of the collected tensors. Default: ""

    Returns
    -------
    * res: list of torch.Tensor
        Sequence of tensors collected recursively in the object.
    * name: list of str
        Sequence of names of the collected tensors.
    """

    # get the tensors recursively towards torch.nn.Module
    res = []
    names = []
    def action(elmt, name, objdict, key):
        res.append(elmt)
        names.append(name)

    # traverse down the object to collect the tensors
    crit = lambda elmt: isinstance(elmt, torch.Tensor) and elmt.dtype in torch_float_type
    _traverse_obj(obj, action=action, crit=crit, prefix=prefix, max_depth=max_depth)
    return res, names

def _set_tensors(obj, all_params, max_depth=20):
    """
    Set the tensors in an object to new tensor object listed in `all_params`.

    Arguments
    ---------
    * obj: an instance
        The object user wants to traverse down
    * all_params: list of torch.Tensor
        Sequence of tensors to be put in the object.
    * max_depth: int
        Maximum recursive depth to avoid infinitely running program.
        If the maximum depth is reached, then raise a RecursionError.
    """
    def action(elmt, name, objdict, key):
        objdict[key] = all_params.pop(0)
    # traverse down the object to collect the tensors
    crit = lambda elmt: isinstance(elmt, torch.Tensor) and elmt.dtype in torch_float_type
    _traverse_obj(obj, action=action, crit=crit, prefix="", max_depth=max_depth)



import torch

# All functions in this file should have the following inputs and outputs
# Inputs
# ------
# * fcn: callable dy/dt = fcn(t, y, *params)
#       The function to be integrated. It should produce output of list of
#       tensors following the shapes of tuple `y`. `t` should be a single element.
# * t: torch.Tensor (nt,)
#       The integrated times
# * y0: list of torch.Tensor (*ny)
#       The list of initial values
# * params: list
#       List of any other parameters
# * **kwargs: dict
#       Any other keyword arguments
# Outputs
# -------
# * yt: list of torch.Tensor (nt,*ny)
#       The value of `y` at the given time `t`
# Note
# ----
# The operations are done in grad-disabled environment and **not** expected to
# be able to propagate gradients.

__all__ = ["rk4_ivp", "rk38_ivp"]

def explicit_rk(tableau, fcn, t, y0, params):
    c = tableau["c"]
    a = tableau["a"]
    b = tableau["b"]
    s = len(c)
    nt = len(t)
    dtype = t.dtype
    device = t.device

    # set up the results list
    yt = torch.empty((nt, *y0.shape), dtype=dtype, device=device)

    yt[0] = y0
    y = y0
    # see https://en.wikipedia.org/wiki/Runge%E2%80%93Kutta_methods#Explicit_Runge.E2.80.93Kutta_methods
    # for the implementation
    for i in range(nt-1):
        t0 = t[i]
        t1 = t[i+1]
        h = t1 - t0
        ks = []
        ksum = 0.0
        for j in range(s):
            if j == 0:
                k = fcn(t0, y, *params)
            else:
                ak = 0.0
                for m in range(j):
                    ak = a[j][m] * ks[m] + ak
                k = fcn(t0 + c[j] * h, h * ak + y, *params)
            ks.append(k)
            ksum += b[j] * k
        y = h * ksum + y
        yt[i+1] = y
    return yt

############################# list of tableaus #############################
rk4_tableau = {
    "c": [0.0, 0.5, 0.5, 1.0],
    "b": [1/6., 1/3., 1/3., 1/6.],
    "a": [[0.0, 0.0, 0.0, 0.0],
          [0.5, 0.0, 0.0, 0.0],
          [0.0, 0.5, 0.0, 0.0],
          [0.0, 0.0, 1.0, 0.0]]
}
rk38_tableau = {
    "c": [0.0, 1/3, 2/3, 1.0],
    "b": [1/8, 3/8, 3/8, 1/8],
    "a": [[ 0.0, 0.0, 0.0, 0.0],
          [ 1/3, 0.0, 0.0, 0.0],
          [-1/3, 1.0, 0.0, 0.0],
          [ 1.0,-1.0, 1.0, 0.0]]
}

############################# list of methods #############################
def rk38_ivp(fcn, t, y0, params, **kwargs):
    return explicit_rk(rk38_tableau, fcn, t, y0, params)

# explicit rk4 implementation to speed up
def rk4_ivp(fcn, t, y0, params, **kwargs):
    """
    Perform the Runge-Kutta steps of order 4 with a fixed step size.
    """
    dtype = t.dtype
    device = t.device
    nt = torch.numel(t)

    # set up the results
    yt = torch.empty((nt, *y0.shape), dtype=dtype, device=device)

    yt[0] = y0
    y = y0
    for i in range(nt-1):
        t0 = t[i]
        t1 = t[i+1]
        h = t1 - t0
        h2 = h * 0.5
        k1 = fcn(t0, y, *params)
        k2 = fcn(t0 + h2, h2 * k1 + y, *params)
        k3 = fcn(t0 + h2, h2 * k2 + y, *params)
        k4 = fcn(t0 + h , h  * k3 + y, *params)
        y = h/6. * (k1 + 2*k2 + 2*k3 + k4) + y
        yt[i+1] = y
    return yt



import torch
import warnings
from typing import Union, Any, Mapping, Optional
import numpy as np
from xitorch import LinearOperator
from xitorch._utils.assertfuncs import assert_runtime
from xitorch._utils.misc import set_default_option, dummy_context_manager
from xitorch._docstr.api_docstr import get_methods_docstr
from xitorch.debug.modes import is_debug_enabled
from xitorch._impls.linalg.solve import exactsolve, wrap_gmres, \
    cg, broyden1_solve, _get_batchdims

def solve(A:LinearOperator, B:torch.Tensor, E:Union[torch.Tensor,None]=None,
          M:Optional[LinearOperator]=None,
          bck_options:Mapping[str,Any]={},
          method:Optional[str]=None,
          **fwd_options) -> torch.Tensor:
    r"""
    Performing iterative method to solve the equation

    .. math::

        \mathbf{AX=B}

    or

    .. math::

        \mathbf{AX-MXE=B}

    where :math:`\mathbf{E}` is a diagonal matrix.
    This function can also solve batched multiple inverse equation at the
    same time by applying :math:`\mathbf{A}` to a tensor :math:`\mathbf{X}`
    with shape ``(...,na,ncols)``.
    The applied :math:`\mathbf{E}` are not necessarily identical for each column.

    Arguments
    ---------
    A: xitorch.LinearOperator
        A linear operator that takes an input ``X`` and produce the vectors in the same
        space as ``B``.
        It should have the shape of ``(*BA, na, na)``
    B: torch.Tensor
        The tensor on the right hand side with shape ``(*BB, na, ncols)``
    E: torch.Tensor or None
        If a tensor, it will solve :math:`\mathbf{AX-MXE = B}`.
        It will be regarded as the diagonal of the matrix.
        Otherwise, it just solves :math:`\mathbf{AX = B}` and ``M`` is ignored.
        If it is a tensor, it should have shape of ``(*BE, ncols)``.
    M: xitorch.LinearOperator or None
        The transformation on the ``E`` side. If ``E`` is ``None``,
        then this argument is ignored.
        If E is not ``None`` and ``M`` is ``None``, then ``M=I``.
        If LinearOperator, it must be Hermitian with shape ``(*BM, na, na)``.
    bck_options: dict
        Options of the iterative solver in the backward calculation.
    method: str or None
        The method of linear equation solver. If ``None``, it will choose
        ``"cg"``.
    **fwd_options
        Method-specific options (see method below)

    Returns
    -------
    torch.Tensor
        The tensor :math:`\mathbf{X}` that satisfies :math:`\mathbf{AX-MXE=B}`.
    """
    assert_runtime(A.shape[-1] == A.shape[-2], "The linear operator A must have a square shape")
    assert_runtime(A.shape[-1] == B.shape[-2], "Mismatch shape of A & B (A: %s, B: %s)" % (A.shape, B.shape))
    if M is not None:
        assert_runtime(M.shape[-1] == M.shape[-2], "The linear operator M must have a square shape")
        assert_runtime(M.shape[-1] == A.shape[-1], "The shape of A & M must match (A: %s, M: %s)" % (A.shape, M.shape))
        assert_runtime(M.is_hermitian, "The linear operator M must be a Hermitian matrix")
    if E is not None:
        assert_runtime(E.shape[-1] == B.shape[-1], "The last dimension of E & B must match (E: %s, B: %s)" % (E.shape, B.shape))
    if E is None and M is not None:
        warnings.warn("M is supplied but will be ignored because E is not supplied")

    # perform expensive check if debug mode is enabled
    if is_debug_enabled():
        A.check()
        if M is not None:
            M.check()

    if method is None:
        method = "cg"

    if method == "exactsolve":
        return exactsolve(A, B, E, M)
    else:
        # get the unique parameters of A
        params = A.getlinopparams()
        mparams = M.getlinopparams() if M is not None else []
        na = len(params)
        return solve_torchfcn.apply(
            A, B, E, M, method,
            fwd_options, bck_options,
            na, *params, *mparams)

class solve_torchfcn(torch.autograd.Function):
    @staticmethod
    def forward(ctx, A, B, E, M, method,
                fwd_options, bck_options,
                na, *all_params):
        # A: (*BA, nr, nr)
        # B: (*BB, nr, ncols)
        # E: (*BE, ncols) or None
        # M: (*BM, nr, nr) or None
        # all_params: list of tensor of any shape
        # returns: (*BABEM, nr, ncols)

        # separate the parameters for A and for M
        params = all_params[:na]
        mparams = all_params[na:]

        config = set_default_option({
        }, fwd_options)
        ctx.bck_config = set_default_option({
        }, bck_options)

        if torch.all(B == 0): # special case
            dims = (*_get_batchdims(A, B, E, M), *B.shape[-2:])
            x = torch.zeros(dims, dtype=B.dtype, device=B.device)
        elif method == "custom_exactsolve":
            x = custom_exactsolve(A, params, B, E=E, M=M, mparams=mparams, **config)
        elif method == "scipy_gmres":
            x = wrap_gmres(A, params, B, E=E, M=M, mparams=mparams, **config)
        elif method == "broyden1":
            x = broyden1_solve(A, params, B, E=E, M=M, mparams=mparams, **config)
        elif method == "cg":
            x = cg(A, params, B, E=E, M=M, mparams=mparams, **config)
        else:
            raise RuntimeError("Unknown solve method: %s" % method)

        ctx.A = A
        ctx.M = M
        ctx.E = E
        ctx.x = x
        ctx.params = params
        ctx.mparams = mparams
        ctx.na = na
        return x

    @staticmethod
    def backward(ctx, grad_x):
        # grad_x: (*BABEM, nr, ncols)
        # ctx.x: (*BABEM, nr, ncols)

        # solve (A-biases*M)^T v = grad_x
        # this is the grad of B
        AT = ctx.A.H # (*BA, nr, nr)
        MT = ctx.M.H if ctx.M is not None else None # (*BM, nr, nr)
        with AT.uselinopparams(*ctx.params), MT.uselinopparams(*ctx.mparams) if MT is not None else dummy_context_manager():
            v = solve(AT, grad_x, ctx.E, MT,
                fwd_options=ctx.bck_config, bck_options=ctx.bck_config) # (*BABEM, nr, ncols)
        grad_B = v

        # calculate the grad of matrices parameters
        with torch.enable_grad():
            params = [p.clone().requires_grad_() for p in ctx.params]
            with ctx.A.uselinopparams(*params):
                loss = -ctx.A.mm(ctx.x) # (*BABEM, nr, ncols)

        grad_params = torch.autograd.grad((loss,), params, grad_outputs=(v,),
            create_graph=torch.is_grad_enabled())

        # calculate the biases gradient
        grad_E = None
        if ctx.E is not None:
            if ctx.M is None:
                Mx = ctx.x
            else:
                with ctx.M.uselinopparams(*ctx.mparams):
                    Mx = ctx.M.mm(ctx.x) # (*BABEM, nr, ncols)
            grad_E = torch.einsum('...rc,...rc->...c', v, Mx) # (*BABEM, ncols)

        # calculate the gradient to the biases matrices
        grad_mparams = []
        if ctx.M is not None and ctx.E is not None:
            with torch.enable_grad():
                mparams = [p.clone().requires_grad_() for p in ctx.mparams]
                lmbdax = ctx.x * ctx.E.unsqueeze(-2)
                with ctx.M.uselinopparams(*mparams):
                    mloss = ctx.M.mm(lmbdax)

            grad_mparams = torch.autograd.grad((mloss,), mparams,
                grad_outputs=(v,),
                create_graph=torch.is_grad_enabled())

        return (None, grad_B, grad_E, None, None, None, None, None,
                *grad_params, *grad_mparams)

def custom_exactsolve(A, params, B, E=None,
                M=None, mparams=[], **options):
    # A: (*BA, na, na)
    # B: (*BB, na, ncols)
    # E: (*BE, ncols)
    # M: (*BM, na, na)
    with A.uselinopparams(*params), M.uselinopparams(*mparams) if M is not None else dummy_context_manager():
        return exactsolve(A, B, E, M)

# docstring completion
_solve_methods = {
    "cg": cg,
    "exactsolve": exactsolve,
    "broyden1": broyden1_solve,
    "scipy_gmres": wrap_gmres,
}
ignore_kwargs = ["E", "M", "mparams"]
solve.__doc__ = get_methods_docstr(solve, _solve_methods, ignore_kwargs)



# file mostly from SciPy: https://github.com/scipy/scipy/blob/914523af3bc03fe7bf61f621363fca27e97ca1d6/scipy/optimize/nonlin.py#L221
# and converted to PyTorch for GPU efficiency

import torch
import functools
from xitorch._impls.optimize.root._jacobian import BroydenFirst

__all__ = ["broyden1"]

def _nonlin_solver(fcn, x0, params, method,
        # jacobian parameters
        alpha=None, max_rank=None,
        # stopping criteria
        maxiter=None, f_tol=None, f_rtol=None, x_tol=None, x_rtol=None,
        # algorithm parameters
        line_search=True,
        # misc parameters
        verbose=False,
        **unused):
    """
    Keyword arguments
    -----------------
    alpha: float or None
        The initial guess of Jacobian is ``-1/alpha``
    max_rank: int or None
        The maximum rank of inverse Jacobian approximation. If ``None``, it
        is ``inf``.
    maxiter: int or None
        Maximum number of iterations, or inf if it is set to None.
    f_tol: float or None
        The absolute tolerance of the norm of the output ``f``.
    f_rtol: float or None
        The relative tolerance of the norm of the output ``f``.
    x_tol: float or None
        The absolute tolerance of the norm of the input ``x``.
    x_rtol: float or None
        The relative tolerance of the norm of the input ``x``.
    line_search: bool or str
        Options to perform line search. If ``True``, it is set to ``"armijo"``.
    verbose: bool
        Options for verbosity
    """

    stop_cond = TerminationCondition(f_tol, f_rtol, x_tol, x_rtol)
    jacobian = {
        "broyden1": BroydenFirst,
    }[method](alpha, max_rank)

    if maxiter is None:
        maxiter = 100*(torch.numel(x0)+1)
    if line_search is True:
        line_search = "armijo"
    elif line_search is False:
        line_search = None

    # shorthand for the function
    xshape = x0.shape
    func = lambda x: fcn(x.reshape(xshape), *params).reshape(-1)
    x = x0.reshape(-1)

    y = func(x)
    y_norm = y.norm()
    if (y_norm == 0):
        return x.reshape(xshape)

    # set up the jacobian
    jacobian.setup(x, y)

    # solver tolerance
    gamma = 0.9
    eta_max = 0.9999
    eta_threshold = 0.1
    eta = 1e-3

    converge = False
    for i in range(maxiter):
        tol = min(eta, eta * y_norm)
        dx = -jacobian.solve(y, tol=tol)

        dx_norm = dx.norm()
        if dx_norm == 0:
            raise ValueError("Jacobian inversion yielded zero vector. "
                             "This indicates a bug in the Jacobian "
                             "approximation.")

        if line_search:
            s, xnew, ynew, y_norm_new = _nonline_line_search(func, x, y, dx,
                search_type=line_search)
        else:
            s = 1.0
            xnew = x + dx
            ynew = func(xnew)
            y_norm_new = ynew.norm()

        jacobian.update(xnew.clone(), ynew)

        # print out dx and df
        dy = (ynew - y)
        to_stop = stop_cond.check(x.norm(), y_norm, dx_norm, dy.norm())
        if verbose:
            if i < 10 or i % 10 == 0 or to_stop:
                print("%6d: |dx|=%.3e, |df|=%.3e" % (i, dx_norm, dy.norm()))
        if to_stop:
            converge = True
            break

        # adjust forcing parameters for inexact solve
        eta_A = float(gamma * (y_norm_new / y_norm)**2)
        gamma_eta2 = gamma * eta * eta
        if gamma_eta2 < eta_threshold:
            eta = min(eta_max, eta_A)
        else:
            eta = min(eta_max, max(eta_A, gamma_eta2))

        y_norm = y_norm_new
        x = xnew
        y = ynew
    if not converge:
        msg = "The rootfinder does not converge after %d iterations. " + \
              "|dx|=%.3e, |df|=%.3e"
        warnings.warn(msg % (maxiter, dx_norm, dy.norm()))
    return x.reshape(xshape)

@functools.wraps(_nonlin_solver, assigned=('__annotations__',)) # takes only the signature
def broyden1(fcn, x0, params=(), **kwargs):
    """
    Solve the root finder or linear equation using the first Broyden method [1]_.
    It can be used to solve minimization by finding the root of the
    function's gradient.

    References
    ----------
    .. [1] B.A. van der Rotten, PhD thesis,
           "A limited memory Broyden method to solve high-dimensional systems of nonlinear equations".
           Mathematisch Instituut, Universiteit Leiden, The Netherlands (2003).
           https://web.archive.org/web/20161022015821/http://www.math.leidenuniv.nl/scripties/Rotten.pdf
    """
    return _nonlin_solver(fcn, x0, params, "broyden1", **kwargs)

# set the docstring of the functions
broyden1.__doc__ += _nonlin_solver.__doc__  # type: ignore

def _safe_norm(v):
    if not torch.isfinite(v).all():
        return torch.tensor(float("inf"), dtype=v.dtype, device=v.device)
    return torch.norm(v)

def _nonline_line_search(func, x, y, dx, search_type="armijo", rdiff=1e-8, smin=1e-2):
    tmp_s = [0]
    tmp_y = [y]
    tmp_phi = [y.norm()**2]
    s_norm = x.norm() / dx.norm()

    def phi(s, store=True):
        if s == tmp_s[0]:
            return tmp_phi[0]
        xt = x + s*dx
        v = func(xt)
        p = _safe_norm(v)**2
        if store:
            tmp_s[0] = s
            tmp_phi[0] = p
            tmp_y[0] = v
        return p

    def derphi(s):
        ds = (torch.abs(s) + s_norm + 1) * rdiff
        return (phi(s+ds, store=False) - phi(s)) / ds

    if search_type == 'armijo':
        s, phi1 = _scalar_search_armijo(phi, tmp_phi[0], -tmp_phi[0],
                                        amin=smin)

    if s is None:
        # No suitable step length found. Take the full Newton step,
        # and hope for the best.
        s = 1.0

    x = x + s*dx
    if s == tmp_s[0]:
        y = tmp_y[0]
    else:
        y = func(x)
    y_norm = y.norm()

    return s, x, y, y_norm

def _scalar_search_armijo(phi, phi0, derphi0, c1=1e-4, alpha0=1, amin=0, max_niter=20):
    phi_a0 = phi(alpha0)
    if phi_a0 <= phi0 + c1*alpha0*derphi0:
        return alpha0, phi_a0

    # Otherwise, compute the minimizer of a quadratic interpolant:

    alpha1 = -(derphi0) * alpha0**2 / 2.0 / (phi_a0 - phi0 - derphi0 * alpha0)
    phi_a1 = phi(alpha1)

    if (phi_a1 <= phi0 + c1*alpha1*derphi0):
        return alpha1, phi_a1

    # Otherwise, loop with cubic interpolation until we find an alpha which
    # satisfies the first Wolfe condition (since we are backtracking, we will
    # assume that the value of alpha is not too small and satisfies the second
    # condition.
    niter = 0
    while alpha1 > amin and niter < max_niter:       # we are assuming alpha>0 is a descent direction
        factor = alpha0**2 * alpha1**2 * (alpha1-alpha0)
        a = alpha0**2 * (phi_a1 - phi0 - derphi0*alpha1) - \
            alpha1**2 * (phi_a0 - phi0 - derphi0*alpha0)
        a = a / factor
        b = -alpha0**3 * (phi_a1 - phi0 - derphi0*alpha1) + \
            alpha1**3 * (phi_a0 - phi0 - derphi0*alpha0)
        b = b / factor

        alpha2 = (-b + torch.sqrt(torch.abs(b**2 - 3 * a * derphi0))) / (3.0*a)
        phi_a2 = phi(alpha2)

        if (phi_a2 <= phi0 + c1*alpha2*derphi0):
            return alpha2, phi_a2

        if (alpha1 - alpha2) > alpha1 / 2.0 or (1 - alpha2/alpha1) < 0.96:
            alpha2 = alpha1 / 2.0

        alpha0 = alpha1
        alpha1 = alpha2
        phi_a0 = phi_a1
        phi_a1 = phi_a2
        niter += 1

    # Failed to find a suitable step length
    if niter == max_niter:
        return alpha2, phi_a2
    return None, phi_a1

class TerminationCondition(object):
    def __init__(self, f_tol, f_rtol, x_tol, x_rtol):
        if f_tol is None:
            f_tol = 1e-6
        if f_rtol is None:
            f_rtol = float("inf")
        if x_tol is None:
            x_tol = 1e-6
        if x_rtol is None:
            x_rtol = float("inf")
        self.f_tol = f_tol
        self.f_rtol = f_rtol
        self.x_tol = x_tol
        self.x_rtol = x_rtol

    def check(self, xnorm, ynorm, dxnorm, dynorm):
        xtcheck = dxnorm < self.x_tol
        xrcheck = dxnorm < self.x_rtol * xnorm
        ytcheck = dynorm < self.f_tol
        yrcheck = dynorm < self.f_rtol * ynorm
        return xtcheck and xrcheck and ytcheck and yrcheck



import torch
from typing import Union, Mapping, Any, Optional, Tuple
import functools
from xitorch import LinearOperator
from xitorch.linalg.solve import solve
from xitorch.debug.modes import is_debug_enabled
from xitorch._utils.assertfuncs import assert_runtime
from xitorch._utils.misc import set_default_option, dummy_context_manager
from xitorch._utils.tensor import ortho
from xitorch._docstr.api_docstr import get_methods_docstr
from xitorch._impls.linalg.symeig import exacteig, davidson

__all__ = ["lsymeig", "usymeig", "symeig", "svd"]

def lsymeig(A:LinearOperator, neig:Optional[int]=None,
        M:Optional[LinearOperator]=None,
        bck_options:Mapping[str,Any]={},
        method:Optional[str]=None,
        **fwd_options) -> Tuple[torch.Tensor, torch.Tensor]:
    return symeig(A, neig, "lowest", M, method=method, bck_options=bck_options, **fwd_options)

def usymeig(A:LinearOperator, neig:Optional[int]=None,
        M:Optional[LinearOperator]=None,
        bck_options:Mapping[str,Any]={},
        method:Optional[str]=None,
        **fwd_options) -> Tuple[torch.Tensor, torch.Tensor]:
    return symeig(A, neig, "uppest", M, method=method, bck_options=bck_options, **fwd_options)

def symeig(A:LinearOperator, neig:Optional[int]=None,
        mode:str="lowest", M:Optional[LinearOperator]=None,
        bck_options:Mapping[str,Any]={},
        method:Optional[str]=None,
        **fwd_options) -> Tuple[torch.Tensor, torch.Tensor]:
    r"""
    Obtain ``neig`` lowest eigenvalues and eigenvectors of a linear operator,

    .. math::

        \mathbf{AX = MXE}

    where :math:`\mathbf{A}, \mathbf{M}` are linear operators,
    :math:`\mathbf{E}` is a diagonal matrix containing the eigenvalues, and
    :math:`\mathbf{X}` is a matrix containing the eigenvectors.

    Arguments
    ---------
    A: xitorch.LinearOperator
        The linear operator object on which the eigenpairs are constructed.
        It must be a Hermitian linear operator with shape ``(*BA, q, q)``
    neig: int or None
        The number of eigenpairs to be retrieved. If ``None``, all eigenpairs are
        retrieved
    mode: str
        ``"lowest"`` or ``"uppermost"``/``"uppest"``. If ``"lowest"``,
        it will take the lowest ``neig`` eigenpairs.
        If ``"uppest"``, it will take the uppermost ``neig``.
    M: xitorch.LinearOperator
        The transformation on the right hand side. If ``None``, then ``M=I``.
        If specified, it must be a Hermitian with shape ``(*BM, q, q)``.
    bck_options: dict
        Method-specific options for :func:`solve` which used in backpropagation
        calculation.
    method: str or None
        Method for the eigendecomposition. If ``None``, it will choose
        ``"exacteig"``.
    **fwd_options
        Method-specific options (see method section below).

    Returns
    -------
    tuple of tensors (eigenvalues, eigenvectors)
        It will return eigenvalues and eigenvectors with shapes respectively
        ``(*BAM, neig)`` and ``(*BAM, na, neig)``, where ``*BAM`` is the
        broadcasted shape of ``*BA`` and ``*BM``.
    """
    assert_runtime(A.is_hermitian, "The linear operator A must be Hermitian")
    if M is not None:
        assert_runtime(M.is_hermitian, "The linear operator M must be Hermitian")
        assert_runtime(M.shape[-1] == A.shape[-1], "The shape of A & M must match (A: %s, M: %s)" % (A.shape, M.shape))
    mode = mode.lower()
    if mode == "uppermost":
        mode = "uppest"
    if method is None: # TODO: do a proper method selection based on size
        method = "exacteig"

    # perform expensive check if debug mode is enabled
    if is_debug_enabled():
        A.check()
        if M is not None:
            M.check()

    if method == "exacteig":
        return exacteig(A, neig, mode, M)
    else:
        fwd_options["method"] = method
        # get the unique parameters of A & M
        params = A.getlinopparams()
        mparams = M.getlinopparams() if M is not None else []
        na = len(params)
        return symeig_torchfcn.apply(A, neig, mode, M,
            fwd_options, bck_options,
            na, *params, *mparams)

def svd(A:LinearOperator, k:Optional[int]=None,
        mode:str="uppest", bck_options:Mapping[str,Any]={},
        method:Optional[str]=None,
        **fwd_options) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    r"""
    Perform the singular value decomposition (SVD):

    .. math::

        \mathbf{A} = \mathbf{U\Sigma V}^H

    where :math:`\mathbf{U}` and :math:`\mathbf{V}` are semi-unitary matrix and
    :math:`\mathbf{\Sigma}` is a diagonal matrix containing real non-negative
    numbers.

    Arguments
    ---------
    A: xitorch.LinearOperator
        The linear operator to be decomposed. It has a shape of ``(*BA, m, n)``
        where ``(*BA)`` is the batched dimension of ``A``.
    k: int or None
        The number of decomposition obtained. If ``None``, it will be
        ``min(*A.shape[-2:])``
    mode: str
        ``"lowest"`` or ``"uppermost"``/``"uppest"``. If ``"lowest"``,
        it will take the lowest ``k`` decomposition.
        If ``"uppest"``, it will take the uppermost ``k``.
    bck_options: dict
        Method-specific options for :func:`solve` which used in backpropagation
        calculation.
    method: str or None
        Method for the svd (same options for :func:`symeig`). If ``None``,
        it will choose ``"exacteig"``.
    **fwd_options
        Method-specific options (see method section below).

    Returns
    -------
    tuple of tensors (u, s, vh)
        It will return ``u, s, vh`` with shapes respectively
        ``(*BA, m, k)``, ``(*BA, k)``, and ``(*BA, k, n)``.

    Note
    ----
    It is a naive implementation of symmetric eigendecomposition of ``A.H @ A``
    or ``A @ A.H`` (depending which one is cheaper)

    Warnings
    --------
    * If ``s`` contains very small numbers or degenerate values, the
      calculation and its gradient might be inaccurate.
    * The second derivative through U or V might be unstable.
      Extra care must be taken.
    """
    # A: (*BA, m, n)
    # adapted from scipy.sparse.linalg.svds

    if is_debug_enabled():
        A.check()
    BA = A.shape[:-2]

    m = A.shape[-2]
    n = A.shape[-1]
    if m < n:
        AAsym = A.matmul(A.H, is_hermitian=True)
        min_nm = m
    else:
        AAsym = A.H.matmul(A, is_hermitian=True)
        min_nm = n

    eivals, eivecs = symeig(AAsym, k, mode,
        bck_options=bck_options, method=method,
        **fwd_options) # (*BA, k) and (*BA, min(mn), k)

    # clamp the eigenvalues to a small positive values to avoid numerical
    # instability
    eivals = torch.clamp(eivals, min=0.0)
    s = torch.sqrt(eivals) # (*BA, k)
    sdiv = torch.clamp(s, min=1e-12).unsqueeze(-2) # (*BA, 1, k)
    if m < n:
        u = eivecs # (*BA, m, k)
        v = A.rmm(u) / sdiv # (*BA, n, k)
    else:
        v = eivecs # (*BA, n, k)
        u = A.mm(v) / sdiv # (*BA, m, k)
    vh = v.transpose(-2, -1)
    return u, s, vh

class symeig_torchfcn(torch.autograd.Function):
    @staticmethod
    def forward(ctx, A, neig, mode, M, fwd_options, bck_options, na, *amparams):
        # A: LinearOperator (*BA, q, q)
        # M: LinearOperator (*BM, q, q) or None

        # separate the sets of parameters
        params = amparams[:na]
        mparams = amparams[na:]

        config = set_default_option({
        }, fwd_options)
        ctx.bck_config = set_default_option({
            # "method": ???
        }, bck_options)

        method = config["method"].lower()
        if method == "davidson":
            evals, evecs = davidson(A, params, neig, mode, M, mparams, **config)
        elif method == "custom_exacteig":
            evals, evecs = custom_exacteig(A, params, neig, mode, M, mparams, **config)
        else:
            raise RuntimeError("Unknown eigen decomposition method: %s" % config["method"])

        # save for the backward
        ctx.evals = evals # (*BAM, neig)
        ctx.evecs = evecs # (*BAM, na, neig)
        ctx.params = params
        ctx.A = A
        ctx.M = M
        ctx.mparams = mparams
        return evals, evecs

    @staticmethod
    def backward(ctx, grad_evals, grad_evecs):
        # grad_evals: (*BAM, neig)
        # grad_evecs: (*BAM, na, neig)

        # get the variables from ctx
        evals = ctx.evals
        evecs = ctx.evecs
        M = ctx.M
        A = ctx.A

        # the loss function where the gradient will be retrieved
        # warnings: if not all params have the connection to the output of A,
        # it could cause an infinite loop because pytorch will keep looking
        # for the *params node and propagate further backward via the `evecs`
        # path. So make sure all the *params are all connected in the graph.
        with torch.enable_grad():
            params = [p.clone().requires_grad_() for p in ctx.params]
            with A.uselinopparams(*params):
                loss = A.mm(evecs) # (*BAM, na, neig)

        # calculate the contributions from the eigenvalues
        gevalsA = grad_evals.unsqueeze(-2) * evecs # (*BAM, na, neig)

        # calculate the contributions from the eigenvectors
        with M.uselinopparams(*ctx.mparams) if M is not None else dummy_context_manager():
            # orthogonalize the grad_evecs with evecs
            B = ortho(grad_evecs, evecs, dim=-2, M=M, mright=False)
            with A.uselinopparams(*ctx.params):
                gevecs = solve(A, -B, evals, M, fwd_options=ctx.bck_config, bck_options=ctx.bck_config)
            # orthogonalize gevecs w.r.t. evecs
            gevecsA = ortho(gevecs, evecs, dim=-2, M=M, mright=True)

        # accummulate the gradient contributions
        gaccumA = gevalsA + gevecsA
        grad_params = torch.autograd.grad(
            outputs=(loss,),
            inputs=params,
            grad_outputs=(gaccumA,),
            create_graph=torch.is_grad_enabled(),
        )

        grad_mparams = []
        if ctx.M is not None:
            with torch.enable_grad():
                mparams = [p.clone().requires_grad_() for p in ctx.mparams]
                with M.uselinopparams(*mparams):
                    mloss = M.mm(evecs) # (*BAM, na, neig)
            gevalsM = -gevalsA * evals.unsqueeze(-2)
            gevecsM = -gevecsA * evals.unsqueeze(-2)

            # the contribution from the parallel elements
            gevecsM_par = (-0.5 * torch.einsum("...ae,...ae->...e", grad_evecs, evecs)).unsqueeze(-2) * evecs # (*BAM, na, neig)

            gaccumM = gevalsM + gevecsM + gevecsM_par
            grad_mparams = torch.autograd.grad(
                outputs=(mloss,),
                inputs=mparams,
                grad_outputs=(gaccumM,),
                create_graph=torch.is_grad_enabled(),
            )

        return (None, None, None, None, None, None, None, *grad_params, *grad_mparams)

def custom_exacteig(A, params, neig, mode, M=None, mparams=[], **options):
    with A.uselinopparams(*params), M.uselinopparams(*mparams) if M is not None else dummy_context_manager():
        return exacteig(A, neig, mode, M)

# docstring completion
_symeig_methods = {
    "exacteig": exacteig,
    "davidson": davidson,
}
ignore_kwargs = ["M", "mparams"]
symeig.__doc__ = get_methods_docstr(symeig, _symeig_methods, ignore_kwargs)
svd.__doc__ = get_methods_docstr(svd, _symeig_methods)



import torch
from xitorch import LinearOperator
from typing import Union
from xitorch._utils.bcast import get_bcasted_dims
from xitorch._utils.misc import dummy_context_manager
from xitorch._utils.tensor import tallqr, to_fortran_order

def exacteig(A:LinearOperator, neig:Union[int,None],
        mode:str, M:Union[LinearOperator,None]):
    """
    Eigendecomposition using explicit matrix construction.
    No additional option for this method.

    Warnings
    --------
    * As this method construct the linear operators explicitly, it might requires
      a large memory.
    """
    Amatrix = A.fullmatrix() # (*BA, q, q)
    if neig is None:
        neig = A.shape[-1]
    if M is None:
        evals, evecs = torch.symeig(Amatrix, eigenvectors=True) # (*BA, q), (*BA, q, q)
        return _take_eigpairs(evals, evecs, neig, mode)
    else:
        Mmatrix = M.fullmatrix() # (*BM, q, q)

        # M decomposition to make A symmetric
        # it is done this way to make it numerically stable in avoiding
        # complex eigenvalues for (near-)degenerate case
        L = torch.cholesky(Mmatrix, upper=False) # (*BM, q, q)
        Linv = torch.inverse(L) # (*BM, q, q)
        LinvT = Linv.transpose(-2,-1) # (*BM, q, q)
        A2 = torch.matmul(Linv, torch.matmul(Amatrix, LinvT)) # (*BAM, q, q)

        # calculate the eigenvalues and eigenvectors
        # (the eigvecs are normalized in M-space)
        evals, evecs = torch.symeig(A2, eigenvectors=True) # (*BAM, q, q)
        evals, evecs = _take_eigpairs(evals, evecs, neig, mode) # (*BAM, neig) and (*BAM, q, neig)
        evecs = torch.matmul(LinvT, evecs)
        return evals, evecs

def davidson(A, params, neig, mode, M=None, mparams=[],
        max_niter=1000,
        nguess=None,
        v_init="randn",
        max_addition=None,
        min_eps=1e-6,
        verbose=False,
        **unused):
    """
    Using Davidson method for large sparse matrix eigendecomposition [1]_.

    Arguments
    ---------
    max_niter: int
        Maximum number of iterations
    nguess: int or None
        The number of initial guess of the eigenvectors
        If None, set to ``neig``.
    v_init: str
        Mode of the initial guess (``"randn"``, ``"rand"``, ``"eye"``)
    max_addition: int or None
        Maximum number of new guesses to be added to the collected vectors.
        If None, set to ``neig``.
    min_eps: float
        Minimum residual error to be stopped
    verbose: bool
        Option to be verbose

    References
    ----------
    .. [1] P. Arbenz, "Lecture Notes on Solving Large Scale Eigenvalue Problems"
           http://people.inf.ethz.ch/arbenz/ewp/Lnotes/chapter12.pdf
    """
    # TODO: optimize for large linear operator and strict min_eps
    # Ideas:
    # (1) use better strategy to get the estimate on eigenvalues
    # (2) use restart strategy

    if nguess is None:
        nguess = neig
    if max_addition is None:
        max_addition = neig

    # get the shape of the transformation
    na = A.shape[-1]
    if M is None:
        bcast_dims = A.shape[:-2]
    else:
        bcast_dims = get_bcasted_dims(A.shape[:-2], M.shape[:-2])
    dtype = A.dtype
    device = A.device

    # TODO: A to use params
    prev_eigvals = None
    prev_eigvalT = None
    stop_reason = "max_niter"
    shift_is_eigvalT = False
    idx = torch.arange(neig).unsqueeze(-1) # (neig, 1)

    with A.uselinopparams(*params), M.uselinopparams(*mparams) if M is not None else dummy_context_manager():
        # set up the initial guess
        V = _set_initial_v(v_init.lower(), dtype, device,
            bcast_dims, na, nguess,
            M=M, mparams=mparams) # (*BAM, na, nguess)
        # V = V.reshape(*bcast_dims, na, nguess) # (*BAM, na, nguess)

        # estimating the lowest eigenvalues
        eig_est, rms_eig = _estimate_eigvals(A, neig, mode,
            bcast_dims=bcast_dims, na=na, ntest=20,
            dtype=V.dtype, device=V.device)

        best_resid = float("inf")
        AV = A.mm(V)
        for i in range(max_niter):
            VT = V.transpose(-2,-1) # (*BAM,nguess,na)
            # Can be optimized by saving AV from the previous iteration and only
            # operate AV for the new V. This works because the old V has already
            # been orthogonalized, so it will stay the same
            # AV = A.mm(V) # (*BAM,na,nguess)
            T = torch.matmul(VT, AV) # (*BAM,nguess,nguess)

            # eigvals are sorted from the lowest
            # eval: (*BAM, nguess), evec: (*BAM, nguess, nguess)
            eigvalT, eigvecT = torch.symeig(T, eigenvectors=True)
            eigvalT, eigvecT = _take_eigpairs(eigvalT, eigvecT, neig, mode) # (*BAM, neig) and (*BAM, nguess, neig)

            # calculate the eigenvectors of A
            eigvecA = torch.matmul(V, eigvecT) # (*BAM, na, neig)

            # calculate the residual
            AVs = torch.matmul(AV, eigvecT) # (*BAM, na, neig)
            LVs = eigvalT.unsqueeze(-2) * eigvecA # (*BAM, na, neig)
            if M is not None:
                LVs = M.mm(LVs)
            resid = AVs - LVs # (*BAM, na, neig)

            # print information and check convergence
            max_resid = resid.abs().max()
            if prev_eigvalT is not None:
                deigval = eigvalT - prev_eigvalT
                max_deigval = deigval.abs().max()
                if verbose:
                    print("Iter %3d (guess size: %d): resid: %.3e, devals: %.3e" % \
                          (i+1, nguess, max_resid, max_deigval))

            if max_resid < best_resid:
                best_resid = max_resid
                best_eigvals = eigvalT
                best_eigvecs = eigvecA
            if max_resid < min_eps:
                break
            if AV.shape[-1] == AV.shape[-2]:
                break
            prev_eigvalT = eigvalT

            # apply the preconditioner
            # initial guess of the eigenvalues are actually help really much
            if not shift_is_eigvalT:
                z = eig_est # (*BAM,neig)
            else:
                z = eigvalT # (*BAM,neig)
            # if A.is_precond_set():
            #     t = A.precond(-resid, *params, biases=z, M=M, mparams=mparams) # (nbatch, na, neig)
            # else:
            t = -resid # (*BAM, na, neig)

            # set the estimate of the eigenvalues
            if not shift_is_eigvalT:
                eigvalT_pred = eigvalT + torch.einsum('...ae,...ae->...e', eigvecA, A.mm(t)) # (*BAM, neig)
                diff_eigvalT = (eigvalT - eigvalT_pred) # (*BAM, neig)
                if diff_eigvalT.abs().max() < rms_eig*1e-2:
                    shift_is_eigvalT = True
                else:
                    change_idx = eig_est > eigvalT
                    next_value = eigvalT - 2*diff_eigvalT
                    eig_est[change_idx] = next_value[change_idx]

            # orthogonalize t with the rest of the V
            t = to_fortran_order(t)
            Vnew = torch.cat((V, t), dim=-1)
            if Vnew.shape[-1] > Vnew.shape[-2]:
                Vnew = Vnew[...,:Vnew.shape[-2]]
            nadd = Vnew.shape[-1] - V.shape[-1]
            nguess = nguess + nadd
            if M is not None:
                MV_ = M.mm(Vnew)
                V, R = tallqr(Vnew, MV=MV_)
            else:
                V, R = tallqr(Vnew)
            AVnew = A.mm(V[...,-nadd:]) # (*BAM,na,nadd)
            AVnew = to_fortran_order(AVnew)
            AV = torch.cat((AV, AVnew), dim=-1)

    eigvals = best_eigvals # (*BAM, neig)
    eigvecs = best_eigvecs # (*BAM, na, neig)
    return eigvals, eigvecs

def _set_initial_v(vinit_type, dtype, device, batch_dims, na, nguess,
                   M=None, mparams=None):
    torch.manual_seed(12421)
    if vinit_type == "eye":
        nbatch = functools.reduce(lambda x,y: x*y, bcast_dims, 1)
        V = torch.eye((na, nguess), dtype=dtype, device=device).unsqueeze(0).repeat(nbatch,1,1).reshape(*batch_dims, na, nguess)
    elif vinit_type == "randn":
        V = torch.randn((*batch_dims, na, nguess), dtype=dtype, device=device)
    elif vinit_type == "random" or vinit_type == "rand":
        V = torch.rand((*batch_dims, na, nguess), dtype=dtype, device=device)
    else:
        raise ValueError("Unknown v_init type: %s" % vinit_type)

    # orthogonalize V
    if M is not None:
        V, R = tallqr(V, MV=M.mm(V))
    else:
        V, R = tallqr(V)
    return V

def _estimate_eigvals(A, neig, mode, bcast_dims, na, ntest, dtype, device):
    # estimate the lowest eigen value
    x = torch.randn((*bcast_dims, na, ntest), dtype=dtype, device=device) # (*BAM, na, ntest)
    x = x / x.norm(dim=-2, keepdim=True)
    Ax = A.mm(x) # (*BAM, na, ntest)
    xTAx = (x * Ax).sum(dim=-2) # (*BAM, ntest)
    mean_eig = xTAx.mean(dim=-1) # (*BAM,)
    std_f = (xTAx).std(dim=-1) # (*BAM,)
    std_x2 = (x*x).std()
    rms_eig = (std_f / std_x2) / (na**0.5) # (*BAM,)
    if mode == "lowest":
        eig_est = mean_eig - 2*rms_eig # (*BAM,)
    else: # uppest
        eig_est = mean_eig + 2*rms_eig # (*BAM,)
    eig_est = eig_est.unsqueeze(-1).repeat_interleave(repeats=neig, dim=-1) # (*BAM,neig)
    return eig_est, rms_eig.max()

def _take_eigpairs(eival, eivec, neig, mode):
    # eival: (*BV, na)
    # eivec: (*BV, na, na)
    if mode == "lowest":
        eival = eival[...,:neig]
        eivec = eivec[...,:neig]
    else: # uppest
        eival = eival[...,-neig:]
        eivec = eivec[...,-neig:]
    return eival, eivec



import warnings
import functools
from typing import Union, Optional, List, Callable, Tuple
import torch
import numpy as np
from xitorch import LinearOperator
from scipy.sparse.linalg import gmres
from xitorch._impls.optimize.root.rootsolver import broyden1
from xitorch._utils.misc import dummy_context_manager
from xitorch._utils.bcast import normalize_bcast_dims, get_bcasted_dims

def wrap_gmres(A, params, B, E=None, M=None, mparams=[],
        min_eps=1e-9,
        max_niter=None,
        **unused):
    """
    Using SciPy's gmres method to solve the linear equation.

    Keyword arguments
    -----------------
    min_eps: float
        Relative tolerance for stopping conditions
    max_niter: int or None
        Maximum number of iterations. If ``None``, default to twice of the
        number of columns of ``A``.
    """
    # A: (*BA, nr, nr)
    # B: (*BB, nr, ncols)
    # E: (*BE, ncols) or None
    # M: (*BM, nr, nr) or None

    # NOTE: currently only works for batched B (1 batch dim), but unbatched A
    assert len(A.shape) == 2 and len(B.shape) == 3, "Currently only works for batched B (1 batch dim), but unbatched A"

    # check the parameters
    msg = "GMRES can only do AX=B"
    assert A.shape[-2] == A.shape[-1], "GMRES can only work for square operator for now"
    assert E is None, msg
    assert M is None, msg

    nbatch, na, ncols = B.shape
    if max_niter is None:
        max_niter = 2*na

    B = B.transpose(-1,-2) # (nbatch, ncols, na)

    # convert the numpy/scipy
    with A.uselinopparams(*params):
        op = A.scipy_linalg_op()
        B_np = B.detach().cpu().numpy()
        res_np = np.empty(B.shape, dtype=np.float64)
        for i in range(nbatch):
            for j in range(ncols):
                x, info = gmres(op, B_np[i,j,:], tol=min_eps, atol=1e-12, maxiter=max_niter)
                if info > 0:
                    msg = "The GMRES iteration does not converge to the desired value "\
                          "(%.3e) after %d iterations" % \
                          (config["min_eps"], info)
                    warnings.warn(msg)
                res_np[i,j,:] = x

        res = torch.tensor(res_np, dtype=B.dtype, device=B.device)
        res = res.transpose(-1,-2) # (nbatch, na, ncols)
        return res

def cg(A:LinearOperator, params:List, B:torch.Tensor,
        E:Optional[torch.Tensor]=None,
        M:Optional[LinearOperator]=None,
        mparams:List=[],
        posdef:bool=False,
        precond:Optional[LinearOperator]=None,
        max_niter:Optional[int]=None,
        rtol:float=1e-6,
        atol:float=1e-8,
        eps:float=1e-12,
        **unused):
    r"""
    Solve the linear equations using Conjugate-Gradient (CG) method.

    Keyword arguments
    -----------------
    posdef: bool
        Indicating if the operation :math:`\mathbf{AX-MXE}` a positive
        definite for all columns and batches.
    precond: LinearOperator or None
        LinearOperator for the preconditioning. If None, no preconditioner is
        applied.
    max_niter: int or None
        Maximum number of iteration. If None, it is set to ``int(1.5 * A.shape[-1])``
    rtol: float
        Relative tolerance for stopping condition w.r.t. norm of B
    atol: float
        Absolute tolerance for stopping condition w.r.t. norm of B
    eps: float
        Substitute the absolute zero in the algorithm's denominator with this
        value to avoid nan.
    """
    def _safedenom(r, eps):
        r[r == 0] = eps
        return r

    def _dot(r, z):
        # r: (*BR, nr, nc)
        # z: (*BR, nr, nc)
        # return: (*BR, 1, nc)
        return torch.einsum("...rc,...rc->...c", r, z).unsqueeze(-2)

    with A.uselinopparams(*params), M.uselinopparams(*mparams) if M is not None else dummy_context_manager():
        nr = A.shape[-1]
        ncols = B.shape[-1]
        if max_niter is None:
            max_niter = int(1.5 * nr)

        # if B is all zeros, then return zeros
        batchdims = _get_batchdims(A, B, E, M)
        if torch.allclose(B, B*0, rtol=rtol, atol=atol):
            x0 = torch.zeros((*batchdims, nr, ncols), dtype=A.dtype, device=A.device)
            return x0

        # setup the preconditioning and the matrix problem
        precond_fcn = _setup_precond(precond)
        A_fcn, B2, col_swapped = _setup_linear_problem(A, B, E, M, posdef)

        # get the stopping matrix
        B_norm = B2.norm(dim=-2, keepdim=True) # (*BB, 1, nc)
        stop_matrix = torch.max(rtol * B_norm, atol * torch.ones_like(B_norm)) # (*BB, 1, nc)

        # prepare the initial guess (it's just all zeros)
        x0shape = (ncols, *batchdims, nr, 1) if col_swapped else (*batchdims, nr, ncols)
        xk = torch.zeros(x0shape, dtype=A.dtype, device=A.device)

        rk = B2 - A_fcn(xk) # (*, nr, nc)
        zk = precond_fcn(rk) # (*, nr, nc)
        pk = zk # (*, nr, nc)
        rkzk = _dot(rk, zk)
        converge = False
        for k in range(max_niter):
            Apk = A_fcn(pk)
            alphak = rkzk / _safedenom(_dot(pk, Apk), eps)
            xk_1 = xk + alphak * pk
            rk_1 = rk - alphak * Apk # (*, nr, nc)

            # check for the stopping condition
            resid = B2 - A_fcn(xk_1)
            resid_norm = resid.norm(dim=-2, keepdim=True)
            if torch.all(resid_norm < stop_matrix):
                converge = True
                break

            zk_1 = precond_fcn(rk_1)
            rkzk_1 = _dot(rk_1, zk_1)
            betak = rkzk_1 / _safedenom(rkzk, eps)
            pk_1 = zk_1 + betak * pk

            # move to the next index
            pk = pk_1
            zk = zk_1
            xk = xk_1
            rk = rk_1
            rkzk = rkzk_1

        if not converge:
            warnings.warn("Convergence is not achieved after %d iterations. "\
                "Max norm of resid: %.3e" % (max_niter, torch.max(resid_norm)))
        if col_swapped:
            # x: (ncols, *, nr, 1)
            xk_1 = xk_1.transpose(0, -1).squeeze(0) # (*, nr, ncols)
        return xk_1

@functools.wraps(broyden1)
def broyden1_solve(A, params, B, E=None, M=None, mparams=[], **options):
    return rootfinder_solve("broyden1", A, params, B, E, M, mparams, **options)

def rootfinder_solve(alg, A, params, B, E=None, M=None, mparams=[], **options):
    # using rootfinder algorithm
    with A.uselinopparams(*params), M.uselinopparams(*mparams) if M is not None else dummy_context_manager():
        nr = A.shape[-1]
        ncols = B.shape[-1]

        # set up the function for the rootfinding
        def fcn_rootfinder(xi):
            # xi: (*BX, nr*ncols)
            x = xi.reshape(*xi.shape[:-1], nr, ncols) # (*BX, nr, ncols)
            y = A.mm(x) - B # (*BX, nr, ncols)
            if E is not None:
                MX = M.mm(x) if M is not None else x
                MXE = MX * E.unsqueeze(-2)
                y = y - MXE # (*BX, nr, ncols)
            y = y.reshape(*xi.shape[:-1], -1) # (*BX, nr*ncols)
            return y

        # setup the initial guess (the batch dimension must be the largest)
        batchdims = _get_batchdims(A, B, E, M)
        x0 = torch.zeros((*batchdims, nr*ncols), dtype=A.dtype, device=A.device)

        if alg == "broyden1":
            x = broyden1(fcn_rootfinder, x0, **options)
        else:
            raise RuntimeError("Unknown method %s" % alg)
        x = x.reshape(*x.shape[:-1], nr, ncols)
        return x

def exactsolve(A:LinearOperator, B:torch.Tensor,
               E:Union[torch.Tensor,None],
               M:Union[LinearOperator,None]):
    """
    Solve the linear equation by contructing the full matrix of LinearOperators.

    Warnings
    --------
    * As this method construct the linear operators explicitly, it might requires
      a large memory.
    """
    # A: (*BA, na, na)
    # B: (*BB, na, ncols)
    # E: (*BE, ncols)
    # M: (*BM, na, na)
    if E is None:
        Amatrix = A.fullmatrix() # (*BA, na, na)
        x, _ = torch.solve(B, Amatrix) # (*BAB, na, ncols)
    elif M is None:
        Amatrix = A.fullmatrix()
        x = _solve_ABE(Amatrix, B, E)
    else:
        Mmatrix = M.fullmatrix() # (*BM, na, na)
        L = torch.cholesky(Mmatrix, upper=False) # (*BM, na, na)
        Linv = torch.inverse(L) # (*BM, na, na)
        LinvT = Linv.transpose(-2,-1) # (*BM, na, na)
        A2 = torch.matmul(Linv, A.mm(LinvT)) # (*BAM, na, na)
        B2 = torch.matmul(Linv, B) # (*BBM, na, ncols)

        X2 = _solve_ABE(A2, B2, E) # (*BABEM, na, ncols)
        x = torch.matmul(LinvT, X2) # (*BABEM, na, ncols)
    return x

def _solve_ABE(A:torch.Tensor, B:torch.Tensor, E:torch.Tensor):
    # A: (*BA, na, na) matrix
    # B: (*BB, na, ncols) matrix
    # E: (*BE, ncols) matrix
    na = A.shape[-1]
    BA, BB, BE = normalize_bcast_dims(A.shape[:-2], B.shape[:-2], E.shape[:-1])
    E = E.reshape(1, *BE, E.shape[-1]).transpose(0,-1) # (ncols, *BE, 1)
    B = B.reshape(1, *BB, *B.shape[-2:]).transpose(0,-1) # (ncols, *BB, na, 1)

    # NOTE: The line below is very inefficient for large na and ncols
    AE = A - torch.diag_embed(E.repeat_interleave(repeats=na, dim=-1), dim1=-2, dim2=-1) # (ncols, *BAE, na, na)
    r, _ = torch.solve(B, AE) # (ncols, *BAEM, na, 1)
    r = r.transpose(0,-1).squeeze(0) # (*BAEM, na, ncols)
    return r

def _get_batchdims(A:LinearOperator, B:torch.Tensor,
        E:Union[torch.Tensor,None],
        M:Union[LinearOperator,None]):

    batchdims = [A.shape[:-2], B.shape[:-2]]
    if E is not None:
        batchdims.append(E.shape[:-1])
        if M is not None:
            batchdims.append(M.shape[:-2])
    return get_bcasted_dims(*batchdims)

def _setup_precond(precond:Optional[LinearOperator]) -> Callable[[torch.Tensor], torch.Tensor]:
    if precond is None:
        precond_fcn = lambda x: x
    elif isinstance(precond, LinearOperator):
        precond_fcn = lambda x: precond.mm(x)
    return precond_fcn

def _setup_linear_problem(A:LinearOperator, B:torch.Tensor,
        E:Optional[torch.Tensor], M:Optional[LinearOperator],
        posdef:bool) -> Tuple[Callable[[torch.Tensor],torch.Tensor], torch.Tensor, bool]:

    # get the linear operator (including the MXE part)
    if E is None:
        A_fcn = lambda x: A.mm(x)
        AT_fcn = lambda x: A.rmm(x)
        B_new = B
        col_swapped = False
    else:
        # A: (*BA, nr, nr) linop
        # B: (*BB, nr, ncols)
        # E: (*BE, ncols)
        # M: (*BM, nr, nr) linop
        if M is None:
            BAs, BBs, BEs = normalize_bcast_dims(A.shape[:-2], B.shape[:-2], E.shape[:-1])
        else:
            BAs, BBs, BEs, BMs = normalize_bcast_dims(A.shape[:-2], B.shape[:-2],
                                                      E.shape[:-1], M.shape[:-2])
        E = E.reshape(*BEs, *E.shape[-1:])
        E_new = E.unsqueeze(0).transpose(-1, 0).unsqueeze(-1) # (ncols, *BEs, 1, 1)
        B = B.reshape(*BBs, *B.shape[-2:]) # (*BBs, nr, ncols)
        B_new = B.unsqueeze(0).transpose(-1, 0) # (ncols, *BBs, nr, 1)

        def A_fcn(x):
            # x: (ncols, *BX, nr, 1)
            Ax = A.mm(x) # (ncols, *BAX, nr, 1)
            Mx = M.mm(x) if M is not None else x # (ncols, *BMX, nr, 1)
            MxE = Mx * E_new # (ncols, *BMXE, nr, 1)
            return Ax - MxE

        def AT_fcn(x):
            # x: (ncols, *BX, nr, 1)
            ATx = A.rmm(x)
            MTx = M.rmm(x) if M is not None else x
            MTxE = MTx * E_new
            return ATx - MTxE

        col_swapped = True

    # get the linear operation if it is not a posdef (A -> AT.A)
    if posdef:
        return A_fcn, B_new, col_swapped
    else:
        def A_new_fcn(x):
            return AT_fcn(A_fcn(x))
        B2 = AT_fcn(B_new)
        return A_new_fcn, B2, col_swapped



import torch
import numpy as np

"""
This file is from https://github.com/locuslab/deq/blob/master/DEQModel/modules/broyden.py
MIT license: https://github.com/locuslab/deq/blob/master/LICENSE
"""

def _safe_norm(v):
    if not torch.isfinite(v).all():
        return np.inf
    return torch.norm(v)

def scalar_search_armijo(phi, phi0, derphi0, c1=1e-4, alpha0=1, amin=0):
    ite = 0
    phi_a0 = phi(alpha0)    # First do an update with step size 1
    if phi_a0 <= phi0 + c1*alpha0*derphi0:
        return alpha0, phi_a0, ite

    # Otherwise, compute the minimizer of a quadratic interpolant
    alpha1 = -(derphi0) * alpha0**2 / 2.0 / (phi_a0 - phi0 - derphi0 * alpha0)
    phi_a1 = phi(alpha1)

    # Otherwise loop with cubic interpolation until we find an alpha which
    # satisfies the first Wolfe condition (since we are backtracking, we will
    # assume that the value of alpha is not too small and satisfies the second
    # condition.
    while alpha1 > amin:       # we are assuming alpha>0 is a descent direction
        factor = alpha0**2 * alpha1**2 * (alpha1-alpha0)
        a = alpha0**2 * (phi_a1 - phi0 - derphi0*alpha1) - \
            alpha1**2 * (phi_a0 - phi0 - derphi0*alpha0)
        a = a / factor
        b = -alpha0**3 * (phi_a1 - phi0 - derphi0*alpha1) + \
            alpha1**3 * (phi_a0 - phi0 - derphi0*alpha0)
        b = b / factor

        alpha2 = (-b + torch.sqrt(torch.abs(b**2 - 3 * a * derphi0))) / (3.0*a)
        phi_a2 = phi(alpha2)
        ite += 1

        if (phi_a2 <= phi0 + c1*alpha2*derphi0):
            return alpha2, phi_a2, ite

        if (alpha1 - alpha2) > alpha1 / 2.0 or (1 - alpha2/alpha1) < 0.96:
            alpha2 = alpha1 / 2.0

        alpha0 = alpha1
        alpha1 = alpha2
        phi_a0 = phi_a1
        phi_a1 = phi_a2

    # Failed to find a suitable step length
    return None, phi_a1, ite

def line_search(update, x0, g0, g, nstep=0, on=True):
    """
    `update` is the propsoed direction of update.
    Code adapted from scipy.
    """
    tmp_s = [0]
    tmp_g0 = [g0]
    tmp_phi = [torch.norm(g0)**2]
    s_norm = torch.norm(x0) / torch.norm(update)

    def phi(s, store=True):
        if s == tmp_s[0]:
            return tmp_phi[0]    # If the step size is so small... just return something
        x_est = x0 + s * update
        g0_new = g(x_est)
        phi_new = _safe_norm(g0_new)**2
        if store:
            tmp_s[0] = s
            tmp_g0[0] = g0_new
            tmp_phi[0] = phi_new
        return phi_new

    if on:
        s, phi1, ite = scalar_search_armijo(phi, tmp_phi[0], -tmp_phi[0], amin=1e-2)
    if (not on) or s is None:
        s = 1.0
        ite = 0

    x_est = x0 + s * update
    if s == tmp_s[0]:
        g0_new = tmp_g0[0]
    else:
        g0_new = g(x_est)
    return x_est - x0, g0_new - g0, ite



import inspect
from typing import Callable, Mapping, Any, Optional, Sequence
import torch
import numpy as np
import scipy.optimize
import xitorch as xt
from xitorch._utils.misc import TensorNonTensorSeparator
from xitorch._utils.assertfuncs import assert_fcn_params
from xitorch._impls.optimize.root.rootsolver import broyden1
from xitorch.linalg.solve import solve
from xitorch.grad.jachess import jac
from xitorch import LinearOperator
from xitorch._core.linop import checklinop
from xitorch._core.editable_module import EditableModule
from xitorch._core.pure_function import get_pure_function, make_sibling
from xitorch._docstr.api_docstr import get_methods_docstr
from xitorch.debug.modes import is_debug_enabled

__all__ = ["equilibrium", "rootfinder", "minimize"]

def rootfinder(
        fcn:Callable[...,torch.Tensor],
        y0:torch.Tensor,
        params:Sequence[Any]=[],
        bck_options:Mapping[str,Any]={},
        method:Optional[str]=None,
        **fwd_options) -> torch.Tensor:
    r"""
    Solving the rootfinder equation of a given function,

    .. math::

        \mathbf{0} = \mathbf{f}(\mathbf{y}, \theta)

    where :math:`\mathbf{f}` is a function that can be non-linear and
    produce output of the same shape of :math:`\mathbf{y}`, and :math:`\theta`
    is other parameters required in the function.
    The output of this block is :math:`\mathbf{y}`
    that produces the :math:`\mathbf{0}` as the output.

    Arguments
    ---------
    fcn : callable
        The function :math:`\mathbf{f}` with output tensor ``(*ny)``
    y0 : torch.tensor
        Initial guess of the solution with shape ``(*ny)``
    params : list
        Sequence of any other parameters to be put in ``fcn``
    bck_options : dict
        Method-specific options for the backward solve (see :func:`xitorch.linalg.solve`)
    method : str or None
        Rootfinder method. If None, it will choose ``"broyden1"``.
    **fwd_options
        Method-specific options (see method section)

    Returns
    -------
    torch.tensor
        The solution which satisfies
        :math:`\mathbf{0} = \mathbf{f}(\mathbf{y},\theta)`
        with shape ``(*ny)``

    Example
    -------
    .. testsetup:: root1

        import torch
        from xitorch.optimize import rootfinder

    .. doctest:: root1

        >>> def func1(y, A):  # example function
        ...     return torch.tanh(A @ y + 0.1) + y / 2.0
        >>> A = torch.tensor([[1.1, 0.4], [0.3, 0.8]]).requires_grad_()
        >>> y0 = torch.zeros((2,1))  # zeros as the initial guess
        >>> yroot = rootfinder(func1, y0, params=(A,))
        >>> print(yroot)
        tensor([[-0.0459],
                [-0.0663]], grad_fn=<_RootFinderBackward>)
    """
    # perform implementation check if debug mode is enabled
    if is_debug_enabled():
        assert_fcn_params(fcn, (y0, *params))

    pfunc = get_pure_function(fcn)
    fwd_options["method"] = _get_rootfinder_default_method(method)
    return _RootFinder.apply(pfunc, y0, fwd_options, bck_options, len(params), *params, *pfunc.objparams())

def equilibrium(
        fcn:Callable[...,torch.Tensor],
        y0:torch.Tensor,
        params:Sequence[Any]=[],
        bck_options:Mapping[str,Any]={},
        method:Optional[str]=None,
        **fwd_options) -> torch.Tensor:
    r"""
    Solving the equilibrium equation of a given function,

    .. math::

        \mathbf{y} = \mathbf{f}(\mathbf{y}, \theta)

    where :math:`\mathbf{f}` is a function that can be non-linear and
    produce output of the same shape of :math:`\mathbf{y}`, and :math:`\theta`
    is other parameters required in the function.
    The output of this block is :math:`\mathbf{y}`
    that produces the same :math:`\mathbf{y}` as the output.

    Arguments
    ---------
    fcn : callable
        The function :math:`\mathbf{f}` with output tensor ``(*ny)``
    y0 : torch.tensor
        Initial guess of the solution with shape ``(*ny)``
    params : list
        Sequence of any other parameters to be put in ``fcn``
    bck_options : dict
        Method-specific options for the backward solve (see :func:`xitorch.linalg.solve`)
    method : str or None
        Rootfinder method. If None, it will choose ``"broyden1"``.
    **fwd_options
        Method-specific options (see method section)

    Returns
    -------
    torch.tensor
        The solution which satisfies
        :math:`\mathbf{y} = \mathbf{f}(\mathbf{y},\theta)`
        with shape ``(*ny)``

    Example
    -------
    .. testsetup:: equil1

        import torch
        from xitorch.optimize import equilibrium

    .. doctest:: equil1

        >>> def func1(y, A):  # example function
        ...     return torch.tanh(A @ y + 0.1) + y / 2.0
        >>> A = torch.tensor([[1.1, 0.4], [0.3, 0.8]]).requires_grad_()
        >>> y0 = torch.zeros((2,1))  # zeros as the initial guess
        >>> yequil = equilibrium(func1, y0, params=(A,))
        >>> print(yequil)
        tensor([[ 0.2313],
                [-0.5957]], grad_fn=<_RootFinderBackward>)
    """
    # perform implementation check if debug mode is enabled
    if is_debug_enabled():
        assert_fcn_params(fcn, (y0, *params))

    pfunc = get_pure_function(fcn)

    @make_sibling(pfunc)
    def new_fcn(y, *params):
        return y - pfunc(y, *params)

    fwd_options["method"] = _get_rootfinder_default_method(method)
    return _RootFinder.apply(new_fcn, y0, fwd_options, bck_options, len(params), *params, *pfunc.getobjparams())

def minimize(
        fcn:Callable[...,torch.Tensor],
        y0:torch.Tensor,
        params:Sequence[Any]=[],
        bck_options:Mapping[str,Any]={},
        method:Optional[str]=None,
        **fwd_options) -> torch.Tensor:
    r"""
    Solve the unbounded minimization problem:

    .. math::

        \mathbf{y^*} = \arg\min_\mathbf{y} f(\mathbf{y}, \theta)

    to find the best :math:`\mathbf{y}` that minimizes the output of the
    function :math:`f`.

    Arguments
    ---------
    fcn: callable
        The function to be optimized with output tensor with 1 element.
    y0: torch.tensor
        Initial guess of the solution with shape ``(*ny)``
    params: list
        Sequence of any other parameters to be put in ``fcn``
    bck_options: dict
        Method-specific options for the backward solve (see :func:`xitorch.linalg.solve`)
    method: str or None
        Minimization method. If None, it will choose ``"broyden1"``.
    **fwd_options
        Method-specific options (see method section)

    Returns
    -------
    torch.tensor
        The solution of the minimization with shape ``(*ny)``

    Example
    -------
    .. testsetup:: root1

        import torch
        from xitorch.optimize import minimize

    .. doctest:: root1

        >>> def func1(y, A):  # example function
        ...     return torch.sum((A @ y)**2 + y / 2.0)
        >>> A = torch.tensor([[1.1, 0.4], [0.3, 0.8]]).requires_grad_()
        >>> y0 = torch.zeros((2,1))  # zeros as the initial guess
        >>> ymin = minimize(func1, y0, params=(A,))
        >>> print(ymin)
        tensor([[-0.0519],
                [-0.2684]], grad_fn=<_RootFinderBackward>)
    """
    # perform implementation check if debug mode is enabled
    if is_debug_enabled():
        assert_fcn_params(fcn, (y0, *params))

    pfunc = get_pure_function(fcn)

    fwd_options["method"] = _get_minimizer_default_method(method)
    # the rootfinder algorithms are designed to move to the opposite direction
    # of the output of the function, so the output of this function is just
    # the grad of z w.r.t. y
    @make_sibling(pfunc)
    def new_fcn(y, *params):
        with torch.enable_grad():
            y1 = y.clone().requires_grad_()
            z = pfunc(y1, *params)
        grady, = torch.autograd.grad(z, (y1,), retain_graph=True,
            create_graph=torch.is_grad_enabled())
        return grady

    return _RootFinder.apply(new_fcn, y0, fwd_options, bck_options, len(params), *params, *pfunc.objparams())

class _RootFinder(torch.autograd.Function):
    @staticmethod
    def forward(ctx, fcn, y0, options, bck_options, nparams, *allparams):

        # set default options
        config = options
        ctx.bck_options = bck_options

        params = allparams[:nparams]
        objparams = allparams[nparams:]

        with fcn.useobjparams(objparams):

            orig_method = config.pop("method")
            method = orig_method.lower()
            if method == "broyden1":
                y = broyden1(fcn, y0, params, **config)
            else:
                raise RuntimeError("Unknown rootfinder method: %s" % orig_method)

        ctx.fcn = fcn

        # split tensors and non-tensors params
        ctx.nparams = nparams
        ctx.param_sep = TensorNonTensorSeparator(allparams)
        tensor_params = ctx.param_sep.get_tensor_params()
        ctx.save_for_backward(y, *tensor_params)

        return y

    @staticmethod
    def backward(ctx, grad_yout):
        param_sep = ctx.param_sep
        yout = ctx.saved_tensors[0]
        nparams = ctx.nparams
        fcn = ctx.fcn

        # merge the tensor and nontensor parameters
        tensor_params = ctx.saved_tensors[1:]
        allparams = param_sep.reconstruct_params(tensor_params)
        params = allparams[:nparams]
        objparams = allparams[nparams:]

        # dL/df
        with fcn.useobjparams(objparams):

            jac_dfdy = jac(ctx.fcn, params=(yout, *params), idxs=[0])[0]
            gyfcn = solve(A=jac_dfdy.H, B=-grad_yout.reshape(-1,1),
                fwd_options=ctx.bck_options, bck_options=ctx.bck_options).reshape(grad_yout.shape)

            # get the grad for the params
            with torch.enable_grad():
                tensor_params_copy = [p.clone().requires_grad_() for p in tensor_params]
                allparams_copy = param_sep.reconstruct_params(tensor_params_copy)
                params_copy = allparams_copy[:nparams]
                objparams_copy = allparams_copy[nparams:]
                with fcn.useobjparams(objparams_copy):
                    yfcn = fcn(yout, *params_copy)

            grad_tensor_params = torch.autograd.grad(yfcn, tensor_params_copy, grad_outputs=gyfcn,
                create_graph=torch.is_grad_enabled())
            grad_nontensor_params = [None for _ in range(param_sep.nnontensors())]
            grad_params = param_sep.reconstruct_params(grad_tensor_params, grad_nontensor_params)

        return (None, None, None, None, None, *grad_params)

def _get_rootfinder_default_method(method):
    if method is None:
        return "broyden1"
    else:
        return method

def _get_minimizer_default_method(method):
    if method is None:
        return "broyden1"
    else:
        return method

# docstring completion
rootfinder.__doc__ = get_methods_docstr(rootfinder, [broyden1])
equilibrium.__doc__ = get_methods_docstr(equilibrium, [broyden1])
minimize.__doc__ = get_methods_docstr(minimize, [broyden1])



def normalize_bcast_dims(*shapes):
    """
    Normalize the lengths of the input shapes to have the same length.
    The shapes are padded at the front by 1 to make the lengths equal.
    """
    maxlens = max([len(shape) for shape in shapes])
    res = [[1]*(maxlens - len(shape)) + list(shape) for shape in shapes]
    return res

def get_bcasted_dims(*shapes):
    """
    Return the broadcasted shape of the given shapes.
    """
    shapes = normalize_bcast_dims(*shapes)
    return [max(*a) for a in zip(*shapes)]



import torch
import numpy as np
import warnings
from xitorch._utils.misc import set_default_option
from xitorch._impls.optimize.linesearch import line_search

def selfconsistent(f, x0, jinv0=1.0, **options):
    """
    Solve the root finder problem with Broyden's method.

    Arguments
    ---------
    * f: callable
        Callable that takes params as the input and output nfeat-outputs.
    * x0: torch.tensor (nbatch, nfeat)
        Initial value of parameters to be put in the function, f.
    * jinv0: float or torch.tensor (nbatch, nfeat, nfeat)
        The initial inverse of the Jacobian. If float, it will be the diagonal.
    * options: dict or None
        Options of the function.

    Returns
    -------
    * x: torch.tensor (nbatch, nfeat)
        The x that approximate f(x) = 0.
    """
    # set up the default options
    config = set_default_option({
        "max_niter": 20,
        "min_eps": 1e-6,
        "beta": 0.9, # contribution of the new delta_n to the total delta_n
        "jinvdecay": 1.0,
        "decayevery": 100,
        "verbose": False,
    }, options)

    # pull out the options for fast access
    min_eps = config["min_eps"]
    verbose = config["verbose"]
    beta = config["beta"]
    jinvdecay = config["jinvdecay"]
    decayevery = config["decayevery"]

    # pull out the parameters of x0
    nbatch, nfeat = x0.shape
    device = x0.device
    dtype = x0.dtype

    # set up the initial jinv
    jinv = _set_jinv0(jinv0, x0)

    # perform the Broyden iterations
    x = x0
    fx = f(x0) # (nbatch, nfeat)
    stop_reason = "max_niter"
    dx = torch.zeros_like(x).to(x.device)
    bestcrit = float("inf")
    for i in range(config["max_niter"]):
        dxnew = -jinv0 * fx # (nbatch, nfeat)
        dx = (1 - beta) * dx + beta * dxnew
        xnew = x + dx # (nbatch, nfeat)
        fxnew = f(xnew)
        dfnew = fxnew - fx

        # update variables for the next iteration
        fx = fxnew
        x = xnew
        if (i+1) % decayevery == 0:
            jinv = jinv * jinvdecay

        # get the best results
        crit = fx.abs().max()
        if crit < bestcrit:
            bestcrit = crit
            bestx = x

        # check the stopping condition
        if verbose:
            print("Iter %3d: %.3e" % (i+1, crit))
        if torch.allclose(fx, torch.zeros_like(fx), atol=min_eps):
            stop_reason = "min_eps"
            break

    if stop_reason != "min_eps":
        msg = "The selfconsistent iteration does not converge to the required accuracy."
        msg += "\nRequired: %.3e. Achieved: %.3e" % (min_eps, bestcrit)
        warnings.warn(msg)

    return bestx

def diis(f, x0, jinv0=1.0, **options):
    """
    Solve the root finder problem with DIIS method.

    Arguments
    ---------
    * f: callable
        Callable that takes params as the input and output nfeat-outputs.
    * x0: torch.tensor (nbatch, nfeat)
        Initial value of parameters to be put in the function, f.
    * jinv0: float or torch.tensor (nbatch, nfeat, nfeat)
        The initial inverse of the Jacobian. If float, it will be the diagonal.
    * options: dict or None
        Options of the function.

    Returns
    -------
    * x: torch.tensor (nbatch, nfeat)
        The x that approximate f(x) = 0.
    """
    # set up the default options
    config = set_default_option({
        "max_niter": 20,
        "min_eps": 1e-6,
        "max_memory": 20,
        "minit": 10,
        "verbose": False,
    }, options)

    # pull out the options for fast access
    min_eps = config["min_eps"]
    verbose = config["verbose"]
    max_memory = config["max_memory"]
    minit = config["minit"]

    # pull out the parameters of x0
    nbatch, nfeat = x0.shape
    device = x0.device
    dtype = x0.dtype

    # set up the initial jinv
    jinv = _set_jinv0(jinv0, x0)

    # perform the iterations
    x = x0
    fx = f(x0) # (nbatch, nfeat)
    stop_reason = "max_niter"
    bestcrit = float("inf")
    nbatch, nfeat = fx.shape
    x_history = torch.empty((nbatch, max_memory, nfeat), dtype=x.dtype, device=x.device)
    e_history = torch.empty((nbatch, max_memory, nfeat), dtype=x.dtype, device=x.device)
    mfill = 0
    midx = 0
    for i in range(config["max_niter"]):
        if mfill < 2:
            dxnew = -jinv * fx # (nbatch, nfeat)
            xnew = x + dxnew # (nbatch, nfeat)
        else:
            # construct the matrix B
            fx_tensor = e_history[:,:mfill,:] # (nbatch, m, nfeat)
            Bul = torch.matmul(fx_tensor, fx_tensor.transpose(-2,-1)) # (nbatch, m, m)
            Bu = torch.cat((Bul, -torch.ones(Bul.shape[0],Bul.shape[1],1, dtype=Bul.dtype, device=Bul.device)), dim=-1) # (nbatch, m, m+1)
            B = torch.cat((Bu, -torch.ones(Bu.shape[0],1,Bu.shape[-1], dtype=Bu.dtype, device=Bu.device)), dim=-2) # (nbatch, m+1, m+1)
            B[:,-1,-1] = 0.0

            # solve the linear equation to get the coefficients
            a = torch.zeros(B.shape[0], B.shape[1], 1, dtype=B.dtype, device=B.device) # (nbatch, m+1, 1)
            a[:,-1] = -1.0
            c = torch.solve(a, B)[0].squeeze(-1)[:,:mfill] # (nbatch, m)

            x_tensor = x_history[:,:mfill,:] # (nbatch, m, nfeat)
            xnew = (x_tensor * c.unsqueeze(-1)).sum(dim=1) # (nbatch, m)

        fxnew = f(xnew) # (nbatch, nfeat)
        dx = xnew - x

        # update variables for the next iteration
        fx = fxnew
        x = xnew

        # add the history
        if i >= minit:
            x_history[:,midx,:] = x
            # e_history[:,midx,:] = dx
            e_history[:,midx,:] = fx
            midx = (midx + 1) % max_memory
            mfill = (mfill + 1) if mfill < max_memory else mfill

        # get the best results
        crit = fx.abs().max()
        if crit < bestcrit:
            bestcrit = crit
            bestx = x

        # check the stopping condition
        if verbose:
            print("Iter %3d: %.3e" % (i+1, crit))
        if torch.allclose(fx, torch.zeros_like(fx), atol=min_eps):
            stop_reason = "min_eps"
            break

    if stop_reason != "min_eps":
        msg = "The DIIS iteration does not converge to the required accuracy."
        msg += "\nRequired: %.3e. Achieved: %.3e" % (min_eps, bestcrit)
        warnings.warn(msg)

    return bestx

def broyden(f, x0, jinv0=1.0, **options):
    """
    Solve the root finder problem with Broyden's method.

    Arguments
    ---------
    * f: callable
        Callable that takes params as the input and output nfeat-outputs.
    * x0: torch.tensor (nbatch, nfeat)
        Initial value of parameters to be put in the function, f.
    * jinv0: float or torch.tensor (nbatch, nfeat, nfeat)
        The initial inverse of the Jacobian. If float, it will be the diagonal.
    * options: dict or None
        Options of the function.

    Returns
    -------
    * x: torch.tensor (nbatch, nfeat)
        The x that approximate f(x) = 0.
    """
    raise RuntimeError("This method is unfinished. Please use other methods.")

    # set up the default options
    config = set_default_option({
        "max_niter": 20,
        "min_eps": 1e-6,
        "verbose": False,
    }, options)

    # pull out the options for fast access
    min_eps = config["min_eps"]
    verbose = config["verbose"]

    # pull out the parameters of x0
    nbatch, nfeat = x0.shape
    device = x0.device
    dtype = x0.dtype

    # set up the initial jinv
    jinv = _set_jinv0(jinv0, x0)

    # perform the Broyden iterations
    x = x0
    fx = f(x0) # (nbatch, nfeat)
    stop_reason = "max_niter"
    for i in range(config["max_niter"]):
        dxnew = -jinv * fx # (nbatch, nfeat)
        xnew = x + dxnew # (nbatch, nfeat)
        fxnew = f(xnew)
        dfnew = fxnew - fx

        # calculate the new jinv
        xtnew_jinv = torch.bmm(xnew.unsqueeze(1), jinv) # (nbatch, 1, nfeat)
        jinv_dfnew = torch.bmm(jinv, dfnew.unsqueeze(-1)) # (nbatch, nfeat, 1)
        xtnew_jinv_dfnew = torch.bmm(xtnew_jinv, dfnew.unsqueeze(-1)) # (nbatch, 1, 1)
        jinvnew = jinv + torch.bmm(dxnew - jinv_dfnew, xtnew_jinv) / xtnew_jinv_dfnew

        # update variables for the next iteration
        fx = fxnew
        jinv = jinvnew
        x = xnew

        # check the stopping condition
        if verbose:
            print("Iter %3d: %.3e" % (i+1, fx.abs().max()))
        if torch.allclose(fx, torch.zeros_like(fx), atol=min_eps):
            stop_reason = "min_eps"
            break

    if stop_reason != "min_eps":
        msg = "The Broyden iteration does not converge to the required accuracy."
        msg += "\nRequired: %.3e. Achieved: %.3e" % (min_eps, fx.abs().max())
        warnings.warn(msg)

    return x

def lbfgs(f, x0, jinv0=1.0, **options):
    """
    Solve the root finder problem with L-BFGS method.

    Arguments
    ---------
    * f: callable
        Callable that takes params as the input and output nfeat-outputs.
    * x0: torch.tensor (*, nfeat)
        Initial value of parameters to be put in the function, f.
    * jinv0: float or torch.tensor (nbatch, nfeat, nfeat)
        The initial inverse of the Jacobian. If float, it will be the diagonal.
    * options: dict or None
        Options of the function.

    Returns
    -------
    * x: torch.tensor (nbatch, nfeat)
        The x that approximate f(x) = 0.
    """
    config = set_default_option({
        "max_niter": 20,
        "min_eps": 1e-6,
        "max_memory": 10,
        "alpha0": 1.0,
        "linesearch": False,
        "verbose": False,
    }, options)

    # pull out the options for fast access
    min_eps = config["min_eps"]
    max_memory = config["max_memory"]
    verbose = config["verbose"]
    linesearch = config["linesearch"]
    alpha = config["alpha0"]

    # set up the initial jinv and the memories
    H0 = _set_jinv0_diag(jinv0, x0) # (*, nfeat)
    sk_history = []
    yk_history = []
    rk_history = []

    def _apply_Vk(rk, sk, yk, grad):
        # sk: (*, nfeat)
        # yk: (*, nfeat)
        # rk: (*, 1)
        return grad - (sk * grad).sum(dim=-1, keepdim=True) * rk * yk

    def _apply_VkT(rk, sk, yk, grad):
        # sk: (*, nfeat)
        # yk: (*, nfeat)
        # rk: (*, 1)
        return grad - (yk * grad).sum(dim=-1, keepdim=True) * rk * sk

    def _apply_Hk(H0, sk_hist, yk_hist, rk_hist, gk):
        # H0: (*, nfeat)
        # sk: (*, nfeat)
        # yk: (*, nfeat)
        # rk: (*, 1)
        # gk: (*, nfeat)
        nhist = len(sk_hist)
        if nhist == 0:
            return H0 * gk

        k = nhist - 1
        rk = rk_hist[k]
        sk = sk_hist[k]
        yk = yk_hist[k]

        # get the last term (rk * sk * sk.T)
        rksksk = (sk * gk).sum(dim=-1, keepdim=True) * rk * sk

        # calculate the V_(k-1)
        grad = gk
        grad = _apply_Vk(rk_hist[k], sk_hist[k], yk_hist[k], grad)
        grad = _apply_Hk(H0, sk_hist[:k], yk_hist[:k], rk_hist[:k], grad)
        grad = _apply_VkT(rk_hist[k], sk_hist[k], yk_hist[k], grad)
        return grad + rksksk

    def _line_search(xk, gk, dk, g):
        if linesearch:
            dx, dg, nit = line_search(dk, xk, gk, g)
            return xk + dx, gk + dg
        else:
            return xk + alpha*dk, g(xk + alpha*dk)

    # perform the main iteration
    xk = x0
    gk = f(xk)
    bestgk = gk.abs().max()
    bestx = x0
    stop_reason = "max_niter"
    for k in range(config["max_niter"]):
        dk = -_apply_Hk(H0, sk_history, yk_history, rk_history, gk)
        xknew, gknew = _line_search(xk, gk, dk, f)

        # store the history
        sk = xknew - xk # (*, nfeat)
        yk = gknew - gk
        inv_rhok = 1.0 / (sk * yk).sum(dim=-1, keepdim=True) # (*, 1)
        sk_history.append(sk)
        yk_history.append(yk)
        rk_history.append(inv_rhok)
        if len(sk_history) > max_memory:
            sk_history = sk_history[-max_memory:]
            yk_history = yk_history[-max_memory:]
            rk_history = rk_history[-max_memory:]

        # update for the next iteration
        xk = xknew
        # alphakold = alphak
        gk = gknew

        # save the best point
        maxgk = gk.abs().max()
        if maxgk < bestgk:
            bestx = xk
            bestgk = maxgk

        # check the stopping condition
        if verbose:
            print("Iter %3d: %.3e" % (k+1, gk.abs().max()))
        if maxgk < min_eps:
            stop_reason = "min_eps"
            break

    if stop_reason != "min_eps":
        msg = "The L-BFGS iteration does not converge to the required accuracy."
        msg += "\nRequired: %.3e. Achieved: %.3e" % (min_eps, bestgk)
        warnings.warn(msg)

    return bestx

def gradrca(f, x0, jinv0=1.0, **options):
    # set up the default options
    config = set_default_option({
        "max_niter": 20,
        "norders": 2,
        "min_eps": 1e-6,
        "verbose": False,
    }, options)

    # pull out the options for fast access
    min_eps = config["min_eps"]
    verbose = config["verbose"]
    norders = config["norders"]

    # pull out the parameters of x0
    nbatch, nfeat = x0.shape
    device = x0.device
    dtype = x0.dtype

    # set up the initial jinv
    jinv = _set_jinv0(jinv0, x0)

    x = x0
    onesvec = torch.ones_like(x0).unsqueeze(-1).to(x0.device) / np.sqrt(nfeat) # (nbatch, nfeat, 1)
    for i in range(config["max_niter"]):
        xg = x.detach().requires_grad_()
        with torch.enable_grad():
            dx = f(xg)
            vunit = (dx / dx.norm(dim=-1, keepdim=True)).detach() # (nbatch, nfeat)
            loss = (dx * dx).sum(dim=-1)
            derivs = [loss]
            for j in range(norders):
                dldx = torch.autograd.grad(derivs[-1].sum(), (xg,), create_graph=(j<norders-1))[0]
                dldlmbda = (dldx * vunit).sum(dim=-1)
                derivs.append(dldlmbda)

        if norders == 2:
            dstep = (-derivs[1] / derivs[2]) # (nbatch,)
        elif norders == 3:
            dstep = (-derivs[2] + torch.sqrt(derivs[2]*derivs[2] - 2*derivs[1]*derivs[3])) / (derivs[3])
        else:
            raise RuntimeError("Order 4 or higher is not defined.")
        dstep = (jinv * dstep.unsqueeze(-1) * vunit)

        if verbose:
            print("Iter %d: %.3e" % (i+1, dx.detach().abs().max()))

        x = x + dstep
    return x

def _set_jinv0(jinv0, x0):
    nbatch, nfeat = x0.shape
    dtype = x0.dtype
    device = x0.device
    if type(jinv0) == torch.Tensor:
        jinv = jinv0
    else:
        jinv = torch.zeros(1,nfeat).to(dtype).to(device) + jinv0 # (1, nfeat)
    return jinv

def _set_jinv0_diag(jinv0, x0):
    if type(jinv0) == torch.Tensor:
        jinv = jinv0
    else:
        jinv = torch.zeros_like(x0).to(x0.device) + jinv0
    return jinv

if __name__ == "__main__":

    dtype = torch.float
    A = torch.tensor([[[0.9703, 0.1178, 0.5345],
         [0.0629, 0.3352, 0.6431],
         [0.8756, 0.7564, 0.1121]]]).to(dtype)
    xtrue = torch.tensor([[0.8690, 0.4324, 0.9035]]).to(dtype)
    b = torch.bmm(A, xtrue.unsqueeze(-1)).squeeze(-1)

    def f(x):
        return torch.bmm(A, x.unsqueeze(-1)).squeeze(-1) - b

    x0 = torch.zeros_like(xtrue)
    jinv0 = 1.0
    x = lbfgs(f, x0, jinv0, verbose=True)
    # x = broyden(f, x0, jinv0, verbose=True)

    print(A)
    print(x)
    print(xtrue)
    print(f(x))



from typing import Optional, List
import torch
from xitorch._core.editable_module import EditableModule
from xitorch._impls.interpolate.interp_1d import CubicSpline1D
from xitorch._docstr.api_docstr import get_methods_docstr

__all__ = ["Interp1D"]

class Interp1D(EditableModule):
    """
    1D interpolation class. When initializing the class, the `x` must be
    specified and `y` can be specified during initialization or later.

    Arguments
    ---------
    x: torch.Tensor
        The position of known values in tensor with shape ``(nr,)``
    y: torch.Tensor or None
        The values at the given position with shape ``(*BY, nr)``.
        If ``None``, it must be supplied during ``__call__``
    method: str or None
        Interpolation method. If None, it will choose ``"cspline"``.
    **fwd_options
        Method-specific options (see method section below)
    """
    def __init__(self,
            x:torch.Tensor,
            y:Optional[torch.Tensor]=None,
            method:Optional[str]=None,
            **fwd_options):
        if method is None:
            method = "cspline"
        if method == "cspline":
            self.obj = CubicSpline1D(x, y, **fwd_options)
        else:
            raise RuntimeError("Unknown interp1d method: %s" % method)

    def __call__(self, xq:torch.Tensor, y:Optional[torch.Tensor]=None) -> torch.Tensor:
        """
        Arguments
        ----------------
        xq: torch.Tensor
            The position of query points with shape ``(nrq,)``.
        y: torch.Tensor or None
            The values at the given position with shape ``(*BY, nr)``.
            If ``y`` has been specified during ``__init__`` and also
            specified here, the value of ``y`` given here will be ignored.
            If no ``y`` ever specified, then it will raise an error.

        Returns
        -------
        torch.Tensor
            The interpolated values with shape ``(*BY, nrq)``.
        """
        return self.obj(xq, y)

    def getparamnames(self, methodname:str, prefix:str="") -> List[str]:
        """"""
        return [prefix+"obj."+c for c in self.obj.getparamnames()]

# docstring completion
interp1d_methods = {
    "cspline": CubicSpline1D,
}
Interp1D.__doc__ = get_methods_docstr(Interp1D, interp1d_methods)



import torch
import inspect
from typing import Callable, Any, Optional
from xitorch._utils.attr import get_attr, set_attr, del_attr
from xitorch._core.editable_module import EditableModule
from contextlib import contextmanager
from abc import abstractmethod

__all__ = ["get_pure_function", "make_sibling"]

############################ functional ###############################
class PureFunction(object):
    """
    PureFunction class wraps methods to make it stateless and expose the pure
    function to take inputs of the original inputs (`params`) and the object's
    states (`objparams`).
    For functions, this class only acts as a thin wrapper.
    """
    def __init__(self, fcn, fcntocall=None):
        self._fcn = fcn
        self._fcntocall = fcn if fcntocall is None else fcntocall
        self._objparams = self.getobjparams()
        self._nobjparams = len(self._objparams)
        self._state_change_allowed = True

    def objparams(self):
        return self._objparams

    def allparams(self, params):
        return [*params, *self._objparams]

    def __call__(self, *params):
        return self._fcntocall(*params)

    @property
    def fcn(self):
        return self._fcn

    @abstractmethod
    def getobjparams(self):
        pass

    @contextmanager
    @abstractmethod
    def _useobjparams(self, objparams):
        pass

    @contextmanager
    def useobjparams(self, objparams):
        if not self._state_change_allowed:
            raise RuntimeError("The state change is disabled")
        with self._useobjparams(objparams):
            try: yield
            finally: pass

    @contextmanager
    def disable_state_change(self):
        try:
            prev_status = self._state_change_allowed
            self._state_change_allowed = False
            yield
        finally:
            self._state_change_allowed = prev_status

    # NOTE: please refrain from implementing enable_state_change.
    # If you want to enable state change, then just remove
    # `with obj.disable_state_change()` statement in your code

class EditableModulePureFunction(PureFunction):
    def getobjparams(self):
        fcn = self.fcn
        self.obj = fcn.__self__
        self.methodname = fcn.__name__
        objparams = self.obj.getuniqueparams(self.methodname)
        return objparams

    @contextmanager
    def _useobjparams(self, objparams):
        with self.obj.useuniqueparams(self.methodname, *objparams):
            try: yield
            finally: pass

class TorchNNPureFunction(PureFunction):
    def getobjparams(self):
        fcn = self.fcn
        obj = fcn.__self__

        # get the tensors in the torch.nn.Module to be used as params
        named_params = list(obj.named_parameters())
        if len(named_params) == 0:
            paramnames = []
            obj_params = []
        else:
            paramnames, obj_params = zip(*named_params)
        self.names = paramnames
        self.obj = obj
        return list(obj_params)

    @contextmanager
    def _useobjparams(self, objparams):
        nnmodule = self.obj
        names = self.names
        try:
            # substitute the state dictionary of the module with the new tensor

            # save the current state
            state_tensors = [get_attr(nnmodule, name) for name in names]

            # substitute the state with the given tensor
            for (name, param) in zip(names, objparams):
                del_attr(nnmodule, name) # delete require in case the param is not a torch.nn.Parameter
                set_attr(nnmodule, name, param)

            yield nnmodule

        except Exception as exc:
            raise exc
        finally:
            # restore back the saved tensors
            for (name, param) in zip(names, state_tensors):
                set_attr(nnmodule, name, param)

class FunctionPureFunction(PureFunction):
    def getobjparams(self):
        return []

    @contextmanager
    def _useobjparams(self, objparams):
        try: yield
        finally: pass

def make_sibling(pfunc:Callable[...,Any]) -> Callable[[Callable[...,Any]], PureFunction]:
    """
    Used as a decor to mark the decorated function as a sibling method of the
    input ``pfunc``.
    Sibling method is a method that is virtually belong to the same object, but
    behaves differently.
    Changing the state of the decorated function will also change the state of
    ``pfunc`` and its other siblings.
    """
    if not isinstance(pfunc, PureFunction):
        pfunc = _get_pure_function(pfunc)
    def decor(fcn):
        new_pfunc = _get_pure_function(pfunc.fcn, fcntocall=fcn)
        return new_pfunc
    return decor

def get_pure_function(fcn:Callable[...,Any]) -> PureFunction:
    """
    Get the pure function form of the function or method ``fcn``.

    Arguments
    ---------
    fcn: function or method
        Function or method to be converted into a ``PureFunction`` by exposing
        the hidden parameters affecting its outputs.

    Returns
    -------
    PureFunction
        The pure function wrapper
    """
    return _get_pure_function(fcn)

def _get_pure_function(
        fcn,
        fcntocall:Optional[Callable[...,Any]]=None) -> PureFunction:
    if not inspect.ismethod(fcn) and not inspect.isfunction(fcn) and not isinstance(fcn, PureFunction):
        if hasattr(fcn, "__call__"):
            fcn = fcn.__call__
        else:
            raise RuntimeError("The function must be callable")

    if isinstance(fcn, PureFunction):
        pfunc = fcn

    elif inspect.isfunction(fcn):
        pfunc = FunctionPureFunction(fcn, fcntocall=fcntocall)

    # if it is a method from an object, unroll the parameters and add
    # the object's parameters as well
    elif isinstance(fcn.__self__, EditableModule):
        pfunc = EditableModulePureFunction(fcn, fcntocall=fcntocall)

    # do the similar thing as EditableModule for torch.nn.Module, but using
    # obj.parameters() as the substitute as .getparams()
    elif isinstance(fcn.__self__, torch.nn.Module):
        pfunc = TorchNNPureFunction(fcn, fcntocall=fcntocall)

    # return as it is if fcn is just a function and params all are tensors
    else:
        raise RuntimeError("The fcn must be a Python function or a method of torch.nn.Module or xitorch.EditableModule")

    return pfunc



import inspect
import warnings

def deprecated(obj):
    if inspect.ismethod(obj):
        name = "Method %s.%s" % (obj.__class__.__name__, obj.__name__)
    elif inspect.isfunction(obj):
        name = "Function %s" % (obj.__name__)
    elif inspect.isclass(obj):
        name = "Class %s" % (obj.__name__)

    if inspect.ismethod(obj) or inspect.isfunction(obj):
        def fcn(*args, **kwargs):
            warnings.warn("%s is deprecated" % name, stacklevel=2)
            return obj(*args, **kwargs)
        return fcn

    elif inspect.isclass(obj):
        # replace the __init__ function
        old_init = obj.__init__
        def newinit(*args, **kwargs):
            warnings.warn("%s is deprecated" % name, stacklevel=2)
            return old_init(*args, **kwargs)
        obj.__init__ = newinit
        return obj



import os
from setuptools import setup, find_packages

verfile = os.path.abspath(os.path.join("xitorch", "version.py"))
version = {"__file__": verfile}
with open(verfile, "r") as fp:
    exec(fp.read(), version)

setup(
    name='xitorch',
    version=version["get_version"](),
    description='Pytorch-based linear algebra library for large matrix',
    url='https://github.com/mfkasim91/xitorch',
    author='mfkasim91',
    author_email='firman.kasim@gmail.com',
    license='MIT',
    packages=find_packages(),
    python_requires=">=3.6",
    install_requires=[
        "numpy>=1.8.2",
        "scipy>=1.1.0",
        "matplotlib>=1.5.3",
        "torch>=1.5",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Mathematics",
        "License :: OSI Approved :: MIT License",

        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    keywords="project library linear-algebra autograd",
    zip_safe=False
)



import torch

"""
This file contains functions for some linear algebra and basic operations of
torch.tensor.
"""

def tallqr(V, MV=None):
    # faster QR for tall and skinny matrix
    # V: (*BV, na, nguess)
    # MV: (*BM, na, nguess) where M is the basis to make Q M-orthogonal
    # if MV is None, then MV=V
    if MV is None:
        MV = V
    VTV = torch.matmul(V.transpose(-2,-1), MV) # (*BMV, nguess, nguess)
    R = torch.cholesky(VTV, upper=True) # (*BMV, nguess, nguess)
    Rinv = torch.inverse(R) # (*BMV, nguess, nguess)
    Q = torch.matmul(V, Rinv)
    return Q, R

def to_fortran_order(V):
    # V: (...,nrow,ncol)
    # outV: (...,nrow,ncol)

    # check if it is in C-contiguous
    if V.is_contiguous():
        # return V.set_(V.storage(), V.storage_offset(), V.size(), tuple(reversed(V.stride())))
        return V.transpose(-2,-1).contiguous().transpose(-2,-1)
    elif V.transpose(-2,-1).is_contiguous():
        return V
    else:
        raise RuntimeError("Only the last two dimensions can be made Fortran order.")

def ortho(A, B, dim=-2, M=None, mright=True):
    """
    Orthogonalize each column in matrix A w.r.t. matrix B in M basis.
    M is a LinearOperator.
    """
    if M is None:
        return A - (A * B).sum(dim=dim, keepdim=True) * B
    elif mright:
        return A - (M.mm(A) * B).sum(dim=dim, keepdim=True) * B
    else:
        return A - M.mm((A * B).sum(dim=dim, keepdim=True) * B)

def convert_none_grads_to_zeros(grads, inputs):
    is_tuple = isinstance(grads, tuple)
    if is_tuple:
        grads = list(grads)
    for i in range(len(grads)):
        g = grads[i]
        if g is None:
            grads[i] = torch.zeros_like(inputs[i])
    if is_tuple:
        grads = tuple(grads)
    return grads



import torch
from abc import abstractmethod

# taking most of the part from SciPy

class Jacobian(object):
    """
    Base class for the Jacobians used in rootfinder algorithms.
    """
    @abstractmethod
    def setup(self, x0, y0, func):
        pass

    @abstractmethod
    def solve(self, v, tol=0):
        pass

    @abstractmethod
    def update(self, x, y):
        pass

class BroydenFirst(Jacobian):
    def __init__(self, alpha=None, max_rank=None):
        self.alpha = alpha
        self.max_rank = max_rank

    def setup(self, x0, y0):
        self.x_prev = x0
        self.y_prev = y0

        if self.max_rank is None:
            self.max_rank = float('inf')

        if self.alpha is None:
            normy0 = torch.norm(y0)
            ones = torch.ones_like(normy0)
            if normy0:
                self.alpha = 0.5 * torch.max(torch.norm(x0), ones) / normy0
            else:
                self.alpha = ones

        # setup the approximate inverse Jacobian
        self.Gm = LowRankMatrix(-self.alpha, "restart")
        self._reduce = lambda: self.Gm.reduce(self.max_rank)

    def solve(self, v, tol=0):
        res = self.Gm.mv(v)
        return res

    def update(self, x, y):
        dy = y - self.y_prev
        dx = x - self.x_prev
        # update Gm
        self._update(x, y, dx, dy, dx.norm(), dy.norm())

        self.y_prev = y
        self.x_prev = x

    def _update(self, x, y, dx, dy, dxnorm, dynorm):
        # keep the rank small
        self._reduce()

        v = self.Gm.rmv(dx)
        c = dx - self.Gm.mv(dy)
        d = v / torch.dot(dy, v)
        self.Gm = self.Gm.append(c, d)

class LowRankMatrix(object):
    # represents a matrix of `\alpha * I + \sum_n c_n d_n^T`
    def __init__(self, alpha, reduce_method):
        self.alpha = alpha
        self.cns = []
        self.dns = []
        self.reduce_method = {
            "restart": 0,
            "simple": 1
        }[reduce_method]

    def mv(self, v):
        res = self.alpha * v
        for i in range(len(self.dns)):
            res += self.cns[i] * torch.dot(self.dns[i], v)
        return res

    def rmv(self, v):
        res = self.alpha * v
        for i in range(len(self.dns)):
            res += self.dns[i] * torch.dot(self.cns[i], v)
        return res

    def append(self, c, d):
        self.cns.append(c)
        self.dns.append(d)
        if len(self.cns) >= torch.numel(c):
            return FullRankMatrix(self.alpha, self.cns, self.dns)
        return self

    def reduce(self, max_rank, **otherparams):
        if len(self.cns) > max_rank:
            if self.reduce_method == 0: # restart
                del self.cns[:]
                del self.dns[:]
            elif self.reduce_method == 1: # simple
                n = len(self.cns)
                del self.cns[:n-max_rank]
                del self.dns[:n-max_rank]

class FullRankMatrix(object):
    def __init__(self, alpha, cns, dns):
        size = torch.numel(cns[0])
        dtype, device = cns[0].dtype, cns[0].device
        self.mat = torch.eye(size, dtype=dtype, device=device)
        self.mat *= alpha
        for i in range(len(cns)):
            self.mat += torch.ger(cns[i], dns[i])

    def mv(self, v):
        res = torch.matmul(self.mat, v)
        return res

    def rmv(self, v):
        return torch.matmul(self.mat.T, v)

    def append(self, c, d):
        self.mat += torch.ger(c, d)
        return self

    def reduce(self, max_rank, **kwargs):
        pass # ???
